#!/usr/bin/env python
"""
GUI Handler (step 1)
"""

import pyqtgraph as pg
from neutronbraggedge.braggedge import BraggEdge

from ibeatles import DataType, Material
from ibeatles.step1.math_utilities import calculate_delta_lambda
from ibeatles.step1.plot import Step1Plot
from ibeatles.step1.roi import Roi
from ibeatles.utilities.gui_handler import GuiHandler
from ibeatles.utilities.retrieve_data_infos import (
    RetrieveGeneralDataInfos,
    RetrieveGeneralFileInfos,
)


class CustomAxis(pg.AxisItem):
    def tickStrings(self, values, scale, spacing):
        if 0 in values:
            return []
        return ["{:.4f}".format(1.0 / i) for i in values]


class Step1GuiHandler(object):
    def __init__(self, parent=None, data_type=None):
        self.parent = parent

        if data_type is None:
            o_gui = GuiHandler(parent=self.parent)
            data_type = o_gui.get_active_tab()

        self.data_type = data_type

    def initialize_rois_and_labels(self):
        """Reached when ROIs have not been manually initialized but loaded via a session"""
        list_roi = self.parent.list_roi[self.data_type]

        roi_function = None
        if self.data_type == DataType.sample:
            roi_function = self.parent.roi_image_view_changed
            image_view = self.parent.ui.image_view
            method = self.parent.roi_image_view_changed
        elif self.data_type == DataType.ob:
            roi_function = self.parent.roi_ob_image_view_changed
            image_view = self.parent.ui.ob_image_view
            method = self.parent.roi_ob_image_view_changed
        elif self.data_type == DataType.normalized:
            roi_function = self.parent.roi_normalized_image_view_changed
            image_view = self.parent.ui.normalized_image_view
            method = self.parent.roi_normalized_image_view_changed

        already_existing_list_roi = self.parent.list_roi_id[self.data_type]
        already_existing_list_label_roi = self.parent.list_label_roi_id[self.data_type]
        for _roi_id, _label_id in zip(already_existing_list_roi, already_existing_list_label_roi):
            image_view.removeItem(_roi_id)
            image_view.removeItem(_label_id)

        list_roi_id = Roi.setup_roi_id(list_roi=list_roi, roi_function=roi_function)
        self.parent.list_roi_id[self.data_type] = list_roi_id

        list_label_roi_id = []

        for _roi, _roi_id in zip(list_roi, list_roi_id):
            label = _roi[0]
            x0 = int(_roi[1])
            y0 = int(_roi[2])

            # label roi
            label_roi = pg.TextItem(
                html='<div style="text-align: center"><span style="color: #ff0000;">' + label + "</span></div>",
                anchor=(-0.3, 1.3),
                border="w",
                fill=(0, 0, 255, 50),
            )

            # # roi region in image
            # roi = pg.ROI([x0, y0], [width, height])
            # roi.addScaleHandle([1, 1], [0, 0])
            image_view.addItem(_roi_id)
            image_view.addItem(label_roi)
            label_roi.setPos(x0, y0)
            _roi_id.sigRegionChanged.connect(method)

            list_label_roi_id.append(label_roi)

        self.parent.list_label_roi_id[self.data_type] = list_label_roi_id

    def sync_instrument_widgets(self, source="load_data"):
        pass

        # target = 'normalized'
        # if source == 'normalized':
        #     target = 'load_data'
        #
        # list_ui = {'load_data': {'distance': self.parent.ui.distance_source_detector,
        #                          'beam': self.parent.ui.beam_rate,
        #                          'detector': self.parent.ui.detector_offset},
        #            'normalized': {'distance': self.parent.ui.distance_source_detector_2,
        #                           'beam': self.parent.ui.beam_rate_2,
        #                           'detector': self.parent.ui.detector_offset_2}}
        #
        # o_gui = GuiHandler(parent=self.parent)
        # distance_value = o_gui.get_text(ui=list_ui[source]['distance'])
        # detector_value = o_gui.get_text(ui=list_ui[source]['detector'])
        # beam_index = o_gui.get_index_selected(ui=list_ui[source]['beam'])
        #
        # o_gui.set_text(value=distance_value, ui=list_ui[target]['distance'])
        # o_gui.set_text(value=detector_value, ui=list_ui[target]['detector'])
        # o_gui.set_index_selected(index=beam_index, ui=list_ui[target]['beam'])

    def load_data_tab_changed(self, tab_index=0):
        data_type = "sample"

        if tab_index == 0:
            # data_preview_box_label = "Sample Image Preview"
            o_general_infos = RetrieveGeneralFileInfos(parent=self.parent, data_type="sample")
            o_selected_infos = RetrieveGeneralDataInfos(parent=self.parent, data_type="sample")
        else:
            # data_preview_box_label = "Open Beam Image Preview"
            o_general_infos = RetrieveGeneralFileInfos(parent=self.parent, data_type="ob")
            o_selected_infos = RetrieveGeneralDataInfos(parent=self.parent, data_type="ob")
            data_type = "ob"

        o_general_infos.update()
        o_selected_infos.update()

        row_selected = self.row_selected(data_type=data_type)
        data = self.parent.data_metadata[data_type]["data"]
        if len(data) > 0:
            data = data[row_selected]
        o_gui = Step1Plot(parent=self.parent, data_type=data_type, data=data)
        o_gui.all_plots()

    def row_selected(self, data_type="sample"):
        return self.parent.data_metadata[data_type]["list_widget_ui"].currentRow()

    def get_element_selected(self):
        return str(self.parent.ui.list_of_elements.currentText())

    def set_crystal_structure(self, new_crystal_structure):
        nbr_item = self.parent.ui.crystal_structure.count()
        for _row in range(nbr_item):
            _item_of_row = self.parent.ui.crystal_structure.itemText(_row)
            if _item_of_row == new_crystal_structure:
                self.parent.ui.crystal_structure.setCurrentIndex(_row)

    def retrieve_handler_from_local_bragg_edge_list(self, material=None):
        """
        Look if the material is in the local list of Bragg edge and if it is,
        return the dictionary of that material
        """
        if material is None:
            return None

        _local_bragg_edge_list = self.parent.local_bragg_edge_list
        if material in _local_bragg_edge_list.keys():
            return _local_bragg_edge_list[material]

    def add_element_to_local_bragg_edge_list(self, material=None):
        """
        Add a new material into the local bragg edge list
        new entry will be
        'material': {'crystal_structure': '', 'lattice': -1}
        """
        if material is None:
            return None

        o_gui = GuiHandler(parent=self.parent)
        _crystal_structure = o_gui.get_text_selected(ui=self.parent.ui.crystal_structure)
        _lattice = o_gui.get_text(ui=self.parent.ui.lattice_parameter)

        self.parent.local_bragg_edge_list[material] = {
            "crystal_structure": _crystal_structure,
            "lattice": _lattice,
        }

    def update_lattice_and_crystal_when_index_selected(self, fill_lattice_flag=True, fill_crystal_structure_flag=True):
        self.parent.ui.list_of_elements.blockSignals(True)

        _element = self.get_element_selected()

        if _element in self.parent.user_defined_bragg_edge_list.keys():
            if self.parent.user_defined_bragg_edge_list[_element][Material.method_used] == Material.via_d0:
                _lattice = ""
                _crystal_structure = ""

            else:
                _lattice = self.parent.user_defined_bragg_edge_list[_element][Material.lattice]
                _crystal_structure = self.parent.user_defined_bragg_edge_list[_element][Material.crystal_structure]

        else:
            _handler = BraggEdge(material=_element)
            _crystal_structure = _handler.metadata["crystal_structure"][_element]
            _lattice = str(_handler.metadata["lattice"][_element])

        _index = self.parent.ui.list_of_elements.currentIndex()

        if fill_lattice_flag:
            self.parent.ui.lattice_parameter.setText(_lattice)

        if fill_crystal_structure_flag:
            self.set_crystal_structure(_crystal_structure)

        self.parent.ui.list_of_elements.blockSignals(False)

    def select_load_data_row(self, data_type="sample", row=0):
        if data_type == "sample":
            self.parent.ui.list_sample.setCurrentRow(row)
        elif data_type == "ob":
            self.parent.ui.list_open_beam.setCurrentRow(row)
        elif data_type == "normalized":
            self.parent.ui.list_normalized.setCurrentRow(row)

    def update_delta_lambda(self):
        distance_source_detector = float(str(self.parent.ui.distance_source_detector.text()))
        frequency = float(str(self.parent.ui.beam_rate.currentText()))

        delta_lambda = calculate_delta_lambda(distance_source_detector=distance_source_detector, frequency=frequency)

        self.parent.ui.delta_lambda_value.setText("{:.2f}".format(delta_lambda))

    def check_step1_widgets(self):
        if self.parent.data_metadata[self.data_type]["data"].any():
            self.parent.ui.load_data_tab.setTabEnabled(1, True)
            self.parent.ui.tabWidget.setTabEnabled(1, True)

    def check_time_spectra_widgets(self):
        if self.data_type == DataType.normalized:
            data_location = "data_normalized"
        else:
            data_location = "data"
        time_spectra_data = self.parent.data_metadata["time_spectra"][data_location]

        if self.parent.ui.material_display_checkbox.isChecked():
            if len(time_spectra_data) == 0:
                _display_error_label = True
            else:
                _display_error_label = False
        else:
            _display_error_label = False

        self.parent.ui.display_warning.setVisible(_display_error_label)

    def block_instrument_widgets(self, status=True):
        self.parent.ui.detector_offset.blockSignals(status)
        self.parent.ui.distance_source_detector.blockSignals(status)
        self.parent.ui.beam_rate.blockSignals(status)



#!/usr/bin/env python
"""
This module provides a class to export data.
"""

from ibeatles.fitting import FittingKeys
from ibeatles.fitting.kropff import SessionSubKeys
from ibeatles.utilities.array_utilities import from_nparray_to_list
from ibeatles.utilities.json_handler import make_value_json_friendly


@staticmethod
def format_kropff_dict(
    table: dict = None,
    d_dict: dict = None,
    strain_dict: dict = None,
    json_friendly: bool = False,
):
    """

    Parameters
    ----------
    table
    d_dict
    strain_dict
    json_friendly if True, the NaN will be replaced by 'NULL'

    Returns
    -------

    """

    cleaned_table = {}
    for _row in table.keys():
        cleaned_table[_row] = {}
        cleaned_table[_row][FittingKeys.y_axis] = from_nparray_to_list(
            table[_row][FittingKeys.y_axis], json_friendly=json_friendly
        )
        cleaned_table[_row][FittingKeys.x_axis] = from_nparray_to_list(
            table[_row][FittingKeys.x_axis], json_friendly=json_friendly
        )

        cleaned_table[_row][SessionSubKeys.fitted] = {}

        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.high_tof] = {}
        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.high_tof][FittingKeys.x_axis] = from_nparray_to_list(
            table[_row][SessionSubKeys.fitted][SessionSubKeys.high_tof][FittingKeys.x_axis],
            json_friendly=json_friendly,
        )
        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.high_tof][FittingKeys.y_axis] = from_nparray_to_list(
            table[_row][SessionSubKeys.fitted][SessionSubKeys.high_tof][FittingKeys.y_axis],
            json_friendly=json_friendly,
        )

        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.low_tof] = {}
        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.low_tof][FittingKeys.x_axis] = from_nparray_to_list(
            table[_row][SessionSubKeys.fitted][SessionSubKeys.low_tof][FittingKeys.x_axis],
            json_friendly=json_friendly,
        )
        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.low_tof][FittingKeys.y_axis] = from_nparray_to_list(
            table[_row][SessionSubKeys.fitted][SessionSubKeys.low_tof][FittingKeys.y_axis],
            json_friendly=json_friendly,
        )

        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.bragg_peak] = {}
        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.bragg_peak][FittingKeys.x_axis] = (
            from_nparray_to_list(
                table[_row][SessionSubKeys.fitted][SessionSubKeys.bragg_peak][FittingKeys.x_axis],
                json_friendly=json_friendly,
            )
        )
        cleaned_table[_row][SessionSubKeys.fitted][SessionSubKeys.bragg_peak][FittingKeys.y_axis] = (
            from_nparray_to_list(
                table[_row][SessionSubKeys.fitted][SessionSubKeys.bragg_peak][FittingKeys.y_axis],
                json_friendly=json_friendly,
            )
        )

        def format_output(input=None, json_friendly=False):
            """create a json friendly version of the input if required"""
            if json_friendly:
                return make_value_json_friendly(input)
            else:
                return input

        # if json_friendly:
        cleaned_table[_row]["strain"] = {
            "val": format_output(strain_dict[_row]["val"]),
            "err": format_output(strain_dict[_row]["err"]),
        }
        cleaned_table[_row]["d"] = {
            "val": format_output(d_dict[_row]["val"]),
            "err": format_output(d_dict[_row]["err"]),
        }

        cleaned_table[_row]["a0"] = format_output(table[_row]["a0"])
        cleaned_table[_row]["b0"] = format_output(table[_row]["b0"])
        cleaned_table[_row]["ahkl"] = format_output(table[_row]["ahkl"])
        cleaned_table[_row]["bhkl"] = format_output(table[_row]["bhkl"])
        cleaned_table[_row]["tau"] = format_output(table[_row]["tau"])
        cleaned_table[_row]["sigma"] = format_output(table[_row]["sigma"])
        cleaned_table[_row]["lambda_hkl"] = format_output(table[_row]["lambda_hkl"])
        cleaned_table[_row][FittingKeys.row_index] = format_output(table[_row][FittingKeys.row_index])
        cleaned_table[_row][FittingKeys.column_index] = format_output(table[_row][FittingKeys.column_index])

        cleaned_table[_row]["bragg peak threshold"] = format_output(table[_row]["bragg peak threshold"])

        # else:
        #
        #     cleaned_table[_row]['strain'] = {'val': format_output(strain_dict[_row]['val']),
        #                                      'err': format_output(strain_dict[_row]['err'])}
        #     cleaned_table[_row]['d'] = {'val': format_output(d_dict[_row]['val']),
        #                                 'err': format_output(d_dict[_row]['err'])}
        #
        #     cleaned_table[_row]['a0'] = format_output(table[_row]['a0'])
        #     cleaned_table[_row]['b0'] = format_output(table[_row]['b0'])
        #     cleaned_table[_row]['ahkl'] = format_output(table[_row]['ahkl'])
        #     cleaned_table[_row]['bhkl'] = format_output(table[_row]['bhkl'])
        #     cleaned_table[_row]['tau'] = format_output(table[_row]['tau'])
        #     cleaned_table[_row]['sigma'] = format_output(table[_row]['sigma'])
        #     cleaned_table[_row]['lambda_hkl'] = format_output(table[_row]['lambda_hkl'])
        #
        #     cleaned_table[_row]['bragg peak threshold'] = format_output(table[_row]['bragg peak threshold'])

        cleaned_table[_row]["bin_coordinates"] = {
            "x0": format_output(table[_row]["bin_coordinates"]["x0"]),
            "y0": format_output(table[_row]["bin_coordinates"]["y0"]),
            "x1": format_output(table[_row]["bin_coordinates"]["x1"]),
            "y1": format_output(table[_row]["bin_coordinates"]["y1"]),
        }

    return cleaned_table


@staticmethod
def format_kropff_table(table: dict = None, d_dict: dict = None, strain_dict: dict = None):
    formatted_table = [
        "#index, "
        + "bin x0, bin y0, bin x1, bin y1, "
        + "lambda hkl val, lambda hkl err, "
        + "d value, d err, strain, strain error"
    ]
    for _row in table.keys():
        _entry = table[_row]

        _row_index = _row
        _bin_x0 = _entry["bin_coordinates"]["x0"]
        _bin_y0 = _entry["bin_coordinates"]["y0"]
        _bin_x1 = _entry["bin_coordinates"]["x1"]
        _bin_y1 = _entry["bin_coordinates"]["y1"]

        _lambda_hkl_val = _entry["lambda_hkl"]["val"]
        _lambda_hkl_err = _entry["lambda_hkl"]["err"]

        _d_value = d_dict[_row]["val"]
        _d_err = d_dict[_row]["err"]

        _strain_value = strain_dict[_row]["val"]
        _strain_value_err = strain_dict[_row]["err"]

        line = [
            _row_index,
            _bin_x0,
            _bin_y0,
            _bin_y0,
            _bin_y1,
            _lambda_hkl_val,
            _lambda_hkl_err,
            _d_value,
            _d_err,
            _strain_value,
            _strain_value_err,
        ]
        line = [str(_value) for _value in line]

        formatted_table.append(", ".join(line))

    return formatted_table



import numpy as np


def is_int(value):
    is_number = True
    try:
        int(value)
    except ValueError:
        is_number = False

    return is_number


def is_nan(value):
    if np.isnan(value):
        return True

    return False


def is_float(value):
    is_number = True
    try:
        float(value)
    except ValueError:
        is_number = False

    return is_number


def get_random_value(max_value=1):
    _value = np.random.rand()
    return _value * max_value


class MeanRangeCalculation:
    """
    Mean value of all the counts between left_index and right_index
    """

    def __init__(self, data=None):
        self.data = data
        self.nbr_point = len(self.data)

    def calculate_left_right_mean(self, index=-1):
        _data = self.data
        _nbr_point = self.nbr_point

        self.left_mean = np.mean(_data[0 : index + 1])
        self.right_mean = np.mean(_data[index + 1 : _nbr_point])

    def calculate_delta_mean_square(self):
        self.delta_square = np.square(self.left_mean - self.right_mean)


def calculate_inflection_point(data=[]):
    """
    will calculate the inflection point by stepping one data at a time and adding
    all the counts to the left and right of that point. Inflection point will be the max
    of the resulting array
    """
    _working_array = []
    for _index, _y in enumerate(data):
        o_range = MeanRangeCalculation(data=data)
        o_range.calculate_left_right_mean(index=_index)
        o_range.calculate_delta_mean_square()
        _working_array.append(o_range.delta_square)

    _peak_value = _working_array.index(max(_working_array))
    return _peak_value


def get_value_of_closest_match(array_to_look_for=None, value=None, left_margin=True):
    """
    for example if the array_to_look = [1,2,3,4] and the value is 2.3,
    the method will return 2.
    for 2.5, it will return the closest or less than value, in this case 2 again.

    :param array_to_look_for: array where to find the closest match
    :param value: input value to use
    :return: return the closest value to the input value found in the array_to_look
    """
    array_to_look_for_in_float = [float(_value) for _value in array_to_look_for]
    value_float = float(value)

    diff_array = [np.abs(_v - value_float) for _v in array_to_look_for_in_float]
    array_of_matches = np.where(np.min(diff_array) == diff_array)

    if left_margin and len(array_of_matches[0]) == 2:
        index = -1
    else:
        index = 0
    best_match = array_of_matches[0][index]
    return best_match


def get_index_of_closest_match(array_to_look_for=None, value=None, left_margin=True):
    """
    for example if the array_to_look = [1,2,3,4] and the value is 2.3,
    the method will return 2.
    for 2.5, it will return the closest or less than value, in this case 2 again.

    :param array_to_look_for: array where to find the closest match
    :param value: input value to use
    :return: return the closest value to the input value found in the array_to_look
    """
    array_to_look_for_in_float = [float(_value) for _value in array_to_look_for]
    value_float = float(value)

    diff_array = [np.abs(_v - value_float) for _v in array_to_look_for_in_float]
    array_of_matches = np.where(np.min(diff_array) == diff_array)

    if left_margin and len(array_of_matches[0]) == 2:
        index = -1
    else:
        index = 0
    best_match = array_of_matches[0][index]
    return best_match



#!/usr/bin/env python
"""
Export images
"""

import json
import logging
import os
from pathlib import Path

import inflect
import numpy as np
from NeuNorm.normalization import Normalization
from qtpy.QtWidgets import QFileDialog

from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.tools.tof_bin.statistics import Statistics
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.file_handler import FileHandler
from ibeatles.utilities.get import Get
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)


class ExportImages:
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent
        self.logger = logging.getLogger("maverick")

    def run(self):
        working_dir = self.top_parent.session[SessionSubKeys.top_folder]
        _folder = str(
            QFileDialog.getExistingDirectory(
                caption="Select Folder to ExportImages the Images",
                directory=working_dir,
                options=QFileDialog.ShowDirsOnly,
            )
        )

        if _folder == "":
            self.logger.info("User cancel export images!")
            return

        self.logger.info(f"Combined and Binned images will be exported to {_folder}!")

        # # use the combined data array
        # combine_arrays = self.parent.combine_data

        # retrieve the bin currently selected [[0],[1],[2,3],....]
        o_get = Get(parent=self.parent)
        bins_dict = o_get.current_bins_activated()
        number_of_bins = len(bins_dict[TimeSpectraKeys.file_index_array])

        # initialize progress bar
        self.parent.eventProgress.setMinimum(0)
        self.parent.eventProgress.setMaximum(number_of_bins - 1)
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)

        sample_position = self.parent.session[SessionKeys.sample_position]

        file_index_array = bins_dict[TimeSpectraKeys.file_index_array]
        tof_array = bins_dict[TimeSpectraKeys.tof_array]
        lambda_array = bins_dict[TimeSpectraKeys.lambda_array]

        o_statistics = Statistics(parent=self.parent)
        self.logger.info(f"Images will be exported in: {_folder}")

        number_of_file_created = 0
        counts_array = []

        metadata_dict = {"sample_position": sample_position}
        file_info_dict = {}

        for _index, _bin in enumerate(file_index_array):
            self.logger.info(f"bin #: {_bin}")

            if _bin == []:
                self.logger.info("-> empty bin, skipping.")
                self.parent.eventProgress.setValue(_index + 1)
                continue

            short_file_name = f"image_{_index:04d}.tif"
            output_file_name = str(Path(_folder) / short_file_name)
            file_info_dict[short_file_name] = {
                "file_index": _bin,
                "tof": tof_array[_index],
                "lambda": lambda_array[_index],
            }

            self.logger.info(f"-> output_file_name: {output_file_name}")
            number_of_file_created += 1

            # we combine the file listed in _bin using the method
            _data_dict = o_statistics.extract_data_for_this_bin(list_runs=_bin)
            full_image = _data_dict["full_image"]
            counts_array.append(int(np.sum(full_image)))
            o_norm = Normalization()
            o_norm.load(data=full_image)
            o_norm.data["sample"]["file_name"][0] = os.path.basename(output_file_name)
            o_norm.export(folder=_folder, data_type="sample", file_type="tiff")

            self.parent.eventProgress.setValue(_index + 1)

        metadata_dict["files_infos"] = file_info_dict
        metadata_file_name = str(Path(_folder) / "metadata.json")
        with open(metadata_file_name, "w") as json_file:
            json.dump(metadata_dict, json_file)

        # export the new time stamp file
        self.export_time_stamp_file(
            counts_array=counts_array,
            tof_array=self.parent.time_spectra[TimeSpectraKeys.tof_array],
            file_index_array=file_index_array,
            export_folder=_folder,
        )

        self.parent.eventProgress.setVisible(False)
        p = inflect.engine()
        self.logger.info(f"Done exporting {number_of_file_created} " + p.plural("file", number_of_file_created) + "!")
        show_status_message(
            parent=self.parent,
            message=f"ExportImages to folder {_folder} ... Done!",
            status=StatusMessageStatus.ready,
            duration_s=5,
        )

    def export_time_stamp_file(
        self,
        counts_array=None,
        tof_array=None,
        file_index_array=None,
        export_folder=None,
    ):
        """
        modify the time_spectra file to mirror the new bins
        :param counts_array: total counts of the given bin
               tof_array: original tof_array (coming from original spectra file)
               file_index_array: bin index
               export_folder: where to export that new time stamp file
        :return: None
        """
        time_spectra_file_name = str(Path(export_folder) / "image_Spectra.txt")
        new_tof_array = []
        for _list_files_index in file_index_array:
            list_tof = [tof_array[_index] for _index in _list_files_index]
            new_tof_array.append(np.mean(list_tof))

        file_content = ["shutter_time,counts"]
        for _tof, _counts in zip(new_tof_array, counts_array):
            file_content.append(f"{_tof},{_counts}")

        FileHandler.make_ascii_file(data=file_content, output_file_name=time_spectra_file_name)

        self.logger.info(f"Exported the new time spectra file: {time_spectra_file_name} ... Done!")



#!/usr/bin/env python
"""
Reload module
"""

import logging
import os

from ibeatles import DataType
from ibeatles.session import SessionSubKeys
from ibeatles.step1.data_handler import DataHandler
from ibeatles.step1.event_handler import EventHandler as Step1EventHandler
from ibeatles.step3.event_handler import EventHandler as Step3EventHandler
from ibeatles.utilities.file_handler import FileHandler


class Reload:
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent

    def run(self, data_type=DataType.normalized, output_folder=None):
        if data_type == DataType.none:
            return

        list_tiff = FileHandler.get_list_of_tif(folder=output_folder)
        self.top_parent.session_dict[data_type][SessionSubKeys.list_files] = [
            os.path.basename(_file) for _file in list_tiff
        ]
        self.top_parent.session_dict[data_type][SessionSubKeys.current_folder] = os.path.dirname(list_tiff[0])

        if data_type == DataType.sample:
            self._raw_data(list_files=list_tiff, load_data_tab_index=0, data_type=data_type)
        elif data_type == DataType.ob:
            self._raw_data(list_files=list_tiff, load_data_tab_index=1, data_type=data_type)
        elif data_type == DataType.normalized:
            self._normalized_data(list_files=list_tiff)

    def _normalized_data(self, list_files=None):
        """this takes care of loading the files back in the normalized tab"""
        data_type = DataType.normalized
        logging.info(f"Reloading TOF combine data in {data_type}")
        o_load = DataHandler(parent=self.top_parent, data_type=data_type)
        folder = os.path.dirname(list_files[0])
        o_load.import_files_from_folder(folder=folder, extension=[".tiff", ".tif"])
        o_load.import_time_spectra()
        o_event_step3 = Step3EventHandler(parent=self.top_parent, data_type=data_type)
        o_event_step3.update_ui_after_loading_data(folder=folder)
        o_event_step3.check_time_spectra_status()
        self.top_parent.infos_window_update(data_type=data_type)
        self.top_parent.ui.normalized_splitter.setSizes([20, 450])
        self.top_parent.ui.main_tools_tabWidget.setCurrentIndex(1)
        self.top_parent.ui.tabWidget.setCurrentIndex(2)

    def _raw_data(self, list_files=None, load_data_tab_index=0, data_type=DataType.sample):
        """This takes care of loading the files into the appropriate sample or OB tab"""
        logging.info(f"Reloading TOF combine data in {data_type}")
        o_load = DataHandler(parent=self.top_parent, data_type=data_type)
        folder = os.path.dirname(list_files[0])
        o_load.import_files_from_folder(folder=folder, extension=[".tiff", ".tif"])
        o_event_step1 = Step1EventHandler(parent=self.top_parent, data_type=data_type)
        o_event_step1.import_button_clicked_step2(folder=folder)
        self.top_parent.ui.load_data_tab.setCurrentIndex(load_data_tab_index)
        self.top_parent.load_data_tab_changed(tab_index=load_data_tab_index)
        self.top_parent.ui.tabWidget.setCurrentIndex(0)
        self.top_parent.ui.main_tools_tabWidget.setCurrentIndex(0)



import numpy as np
from scipy.ndimage import convolve


def find_nearest_index(array, value):
    idx = (np.abs(np.array(array) - value)).argmin()
    return idx


def get_min_max_xy(pos_array):
    min_x = 10000
    max_x = -1
    min_y = 10000
    max_y = -1

    for xy in pos_array:
        [_x, _y] = xy
        if _x < min_x:
            min_x = _x
        if _x > max_x:
            max_x = _x
        if _y < min_y:
            min_y = _y
        if _y > max_y:
            max_y = _y

    return {"x": {"min": min_x, "max": max_x}, "y": {"min": min_y, "max": max_y}}


def gamma_filtering(data_array, threshold=0.1):
    """
    this algorithm will perform the gamma filtering. That means that
    every pixel counts that are above threshold of the total average counts
    will be replaced by the average value of the 9 pixels around it.
    """

    final_data_array = []
    for _data in data_array:
        _data_filtered = single_gamma_filtering(_data)
        final_data_array.append(_data_filtered)

    return final_data_array


def single_gamma_filtering(data, threshold=0.1):
    raw_data = np.copy(data)

    # find mean counts
    mean_counts = np.mean(raw_data)

    thresolded_raw_data = raw_data * threshold

    # get pixels where value is above threshold
    position = []
    [height, width] = np.shape(raw_data)
    for _x in np.arange(width):
        for _y in np.arange(height):
            if thresolded_raw_data[_y, _x] > mean_counts:
                position.append([_y, _x])

    # convolve entire image using 3x3 kerne
    mean_kernel = np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]]) / 8.0
    convolved_data = convolve(raw_data, mean_kernel, mode="constant")

    # replace only pixel above threshold by convolved data
    for _coordinates in position:
        [_y, _x] = _coordinates
        raw_data[_y, _x] = convolved_data[_y, _x]

    return raw_data


def exclude_y_value_when_error_is_nan(axis, error_axis):
    axis_cleaned = []
    error_axis_cleaned = []

    for _x, _error in zip(axis, error_axis):
        if (_x == "None") or (_error == "None") or (_x is None) or (_error is None):
            axis_cleaned.append(np.nan)
            error_axis_cleaned.append(np.nan)
        else:
            axis_cleaned.append(float(_x))
            error_axis_cleaned.append(float(_error))

    return axis_cleaned, error_axis_cleaned


def calculate_median(array_of_value=None):
    if array_of_value is None:
        return None

    try:
        array_of_value = np.where(np.isinf(array_of_value), np.nan, array_of_value)
        array_of_value = np.where(np.isneginf(array_of_value), np.nan, array_of_value)
    except TypeError:
        return None

    return np.nanmedian(array_of_value)


def from_nparray_to_list(nparray=None, json_friendly=False) -> list:
    if nparray is None:
        return None

    if json_friendly:
        nparray = [str(_value) for _value in nparray]

    return list(nparray)



#!/usr/bin/env python
"""
Get class for handling the Kropff fitting.
"""

import numpy as np

import ibeatles.utilities.error as fitting_error
from ibeatles import DataType
from ibeatles.fitting import FittingKeys, KropffTabSelected
from ibeatles.fitting.kropff import FittingRegions, SessionSubKeys
from ibeatles.utilities.table_handler import TableHandler


class Get:
    image_size = {"height": None, "width": None}

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def a0(self):
        a0 = self.parent.ui.kropff_high_lda_a0_init.text()
        try:
            a0 = float(a0)
        except ValueError:
            raise fitting_error.HighLambdaFittingError(
                fitting_region=FittingRegions.high_lambda,
                message="Wrong a\u2080 format!",
            )
        return a0

    def b0(self):
        b0 = self.parent.ui.kropff_high_lda_b0_init.text()
        try:
            b0 = float(b0)
        except ValueError:
            raise fitting_error.HighLambdaFittingError(
                fitting_region=FittingRegions.high_lambda,
                message="Wrong b\u2080 format!",
            )
        return b0

    def ahkl(self):
        ahkl = self.parent.ui.kropff_low_lda_ahkl_init.text()
        try:
            ahkl = float(ahkl)
        except ValueError:
            raise fitting_error.LowLambdaFittingError(
                fitting_region=FittingRegions.low_lambda,
                message="Wrong a\u2095\u2096\u2097 format!",
            )
        return ahkl

    def bhkl(self):
        bhkl = self.parent.ui.kropff_low_lda_bhkl_init.text()
        try:
            bhkl = float(bhkl)
        except ValueError:
            raise fitting_error.LowLambdaFittingError(
                fitting_region=FittingRegions.low_lambda,
                message="Wrong b\u2095\u2096\u2097 format!",
            )
        return bhkl

    def lambda_hkl(self):
        lambda_hkl = self.parent.kropff_lambda_settings["fix"]
        try:
            lambda_hkl = float(lambda_hkl)
        except ValueError:
            raise fitting_error.BraggPeakFittingError(
                fitting_region=FittingRegions.bragg_peak,
                message="Wrong \u03bb\u2095\u2096\u2097 format!",
            )
        return lambda_hkl

    def tau(self):
        tau = self.parent.ui.kropff_bragg_peak_tau_init.text()
        try:
            tau = float(tau)
        except ValueError:
            raise fitting_error.BraggPeakFittingError(
                fitting_region=FittingRegions.bragg_peak, message="Wrong \u03c4 format!"
            )
        return tau

    def sigma(self):
        sigma = self.parent.ui.kropff_bragg_peak_sigma_comboBox.currentText()
        try:
            sigma = float(sigma)
        except ValueError:
            raise fitting_error.BraggPeakFittingError(
                fitting_region=FittingRegions.bragg_peak, message="Wrong sigma format!"
            )
        return sigma

    def variable_selected(self):
        """get the variable selected in the Check/Set Variables table"""
        if self.parent.ui.lambda_hkl_button.isChecked():
            return SessionSubKeys.lambda_hkl
        elif self.parent.ui.sigma_button.isChecked():
            return SessionSubKeys.sigma
        elif self.parent.ui.tau_button.isChecked():
            return SessionSubKeys.tau
        else:
            raise NotImplementedError("variable requested not supported!")

    def list_lambda_hkl_initial_guess(self):
        return self._list_parameter_initial_guess(
            parameter=SessionSubKeys.lambda_hkl,
            error_message="Wrong \u03bb\u2095\u2096\u2097 format!",
        )

    def list_sigma_initial_guess(self):
        return self._list_parameter_initial_guess(parameter=SessionSubKeys.sigma, error_message="Wrong sigma format!")

    def list_tau_initial_guess(self):
        return self._list_parameter_initial_guess(parameter=SessionSubKeys.tau, error_message="Wrong tau format!")

    def _list_parameter_initial_guess(
        self,
        parameter=SessionSubKeys.lambda_hkl,
        error_message="Wrong \u03bb\u2095\u2096\u2097 format!",
    ):
        list_ui = {
            "fix_value": {
                SessionSubKeys.lambda_hkl: self.parent.ui.lambda_hkl_fix_lineEdit,
                SessionSubKeys.sigma: self.parent.ui.sigma_fix_lineEdit,
                SessionSubKeys.tau: self.parent.ui.tau_fix_lineEdit,
            },
            "fix_radio": {
                SessionSubKeys.lambda_hkl: self.parent.ui.lambda_hkl_fix_radioButton,
                SessionSubKeys.sigma: self.parent.ui.sigma_fix_radioButton,
                SessionSubKeys.tau: self.parent.ui.tau_fix_radioButton,
            },
            "from": {
                SessionSubKeys.lambda_hkl: self.parent.ui.lambda_hkl_from_lineEdit,
                SessionSubKeys.sigma: self.parent.ui.sigma_from_lineEdit,
                SessionSubKeys.tau: self.parent.ui.tau_from_lineEdit,
            },
            "to": {
                SessionSubKeys.lambda_hkl: self.parent.ui.lambda_hkl_to_lineEdit,
                SessionSubKeys.sigma: self.parent.ui.sigma_to_lineEdit,
                SessionSubKeys.tau: self.parent.ui.tau_to_lineEdit,
            },
            "step": {
                SessionSubKeys.lambda_hkl: self.parent.ui.lambda_hkl_step_lineEdit,
                SessionSubKeys.sigma: self.parent.ui.sigma_step_lineEdit,
                SessionSubKeys.tau: self.parent.ui.tau_step_lineEdit,
            },
        }

        try:
            if list_ui["fix_radio"][parameter].isChecked():
                list_value = [float(list_ui["fix_value"][parameter].text())]
            else:
                _from = float(list_ui["from"][parameter].text())
                _to = float(list_ui["to"][parameter].text())
                _step = float(list_ui["step"][parameter].text())

                value = _from
                list_value = []
                while value <= _to:
                    list_value.append(value)
                    value += _step

        except ValueError:
            raise fitting_error.BraggPeakFittingError(fitting_region=FittingRegions.bragg_peak, message=error_message)

        return list_value

    def kropff_row_selected(self):
        kropff_tab_ui_selected = self.kropff_table_ui_selected()
        o_table = TableHandler(table_ui=kropff_tab_ui_selected)
        row_selected = o_table.get_rows_of_table_selected()
        return row_selected

    def kropff_table_ui_selected(self):
        tab_selected = self.kropff_tab_selected()
        if tab_selected == KropffTabSelected.high_tof:
            return self.parent.ui.high_lda_tableWidget
        elif tab_selected == KropffTabSelected.low_tof:
            return self.parent.ui.low_lda_tableWidget
        elif tab_selected == KropffTabSelected.bragg_peak:
            return self.parent.ui.bragg_edge_tableWidget
        elif tab_selected == KropffTabSelected.summary:
            return None
        elif tab_selected == KropffTabSelected.settings:
            return None

    def kropff_tab_selected(self):
        tab_selected_index = self.parent.ui.kropff_tabWidget.currentIndex()
        if tab_selected_index == 1:
            return KropffTabSelected.low_tof
        elif tab_selected_index == 2:
            return KropffTabSelected.high_tof
        elif tab_selected_index == 3:
            return KropffTabSelected.bragg_peak
        elif tab_selected_index == 0:
            return KropffTabSelected.settings
        elif tab_selected_index == 4:
            return KropffTabSelected.summary

    def kropff_matplotlib_ui_selected(self):
        tab_selected = self.kropff_tab_selected()
        if tab_selected == KropffTabSelected.high_tof:
            return self.parent.kropff_high_plot
        elif tab_selected == KropffTabSelected.low_tof:
            return self.parent.kropff_low_plot
        elif tab_selected == KropffTabSelected.bragg_peak:
            return self.parent.kropff_bragg_peak_plot
        elif tab_selected == KropffTabSelected.settings:
            return None
        elif tab_selected == KropffTabSelected.summary:
            return None
        else:
            raise ValueError("Tab selected is invalid!")

    def kropff_fitting_parameters_radioButton_selected(self):
        tab_selected = self.kropff_tab_selected()
        if tab_selected == KropffTabSelected.high_tof:
            if self.parent.ui.kropff_a0_radioButton.isChecked():
                return "a0"
            elif self.parent.ui.kropff_b0_radioButton.isChecked():
                return "b0"
            else:
                raise ValueError("fitting parameters is invalid!")
        elif tab_selected == KropffTabSelected.low_tof:
            if self.parent.ui.kropff_ahkl_radioButton.isChecked():
                return "ahkl"
            elif self.parent.ui.kropff_bhkl_radioButton.isChecked():
                return "bhkl"
            else:
                raise ValueError("fitting parameters is invalid!")
        elif tab_selected == KropffTabSelected.bragg_peak:
            if self.parent.ui.kropff_lda_hkl_radioButton.isChecked():
                return "lambda_hkl"
            elif self.parent.ui.kropff_tau_radioButton.isChecked():
                return "tau"
            elif self.parent.ui.kropff_sigma_radioButton.isChecked():
                return "sigma"
            else:
                raise ValueError("fitting parameters is invalid!")
        elif tab_selected == KropffTabSelected.settings:
            return None
        elif tab_selected == KropffTabSelected.summary:
            return None
        else:
            raise ValueError("Tab selected is invalid!")

    def active_d0(self):
        lambda0 = float(self.parent.ui.bragg_edge_calculated.text())
        return lambda0 / 2.0

    def calculate_image_size(self):
        live_data = self.grand_parent.data_metadata[DataType.normalized]["data"]
        integrated_image = np.mean(live_data, 0)
        self.parent.integrated_image = np.transpose(integrated_image)
        [self.image_size["height"], self.image_size["width"]] = np.shape(integrated_image)

    def calculate_d_array(self):
        self.calculate_image_size()
        width = self.image_size["width"]
        height = self.image_size["height"]

        d_array = np.zeros((height, width))
        # d_array[:] = np.nan
        d_dict = {}

        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        for _row_index in kropff_table_dictionary.keys():
            _row_entry = kropff_table_dictionary[_row_index]

            bin_coordinates = _row_entry["bin_coordinates"]

            x0 = bin_coordinates["x0"]
            x1 = bin_coordinates["x1"]
            y0 = bin_coordinates["y0"]
            y1 = bin_coordinates["y1"]

            lambda_hkl = _row_entry["lambda_hkl"]["val"]
            lambda_hkl_err = _row_entry["lambda_hkl"]["err"]
            if lambda_hkl_err is None:
                lambda_hkl_err = np.sqrt(lambda_hkl)

            d_array[y0:y1, x0:x1] = float(lambda_hkl) / 2.0
            d_dict[_row_index] = {
                "val": float(lambda_hkl) / 2.0,
                "err": float(lambda_hkl_err) / 2.0,
            }

        # self.parent.d_array = d_array
        self.parent.d_dict = d_dict

    def strain_mapping_dictionary(self):
        self.calculate_d_array()
        d_dict = self.parent.d_dict
        strain_mapping_dict = {}
        for _row in d_dict.keys():
            d0 = self.active_d0()
            d = d_dict[_row]["val"]
            d_error = d_dict[_row]["err"]
            strain_mapping = (d - d0) / d0
            strain_mapping_err = d_error + np.sqrt(d0)

            strain_mapping_dict[_row] = {
                "val": strain_mapping,
                "err": strain_mapping_err,
            }

        return strain_mapping_dict

    def nbr_row(self):
        list_row = []
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        for _row_index in kropff_table_dictionary.keys():
            _row = kropff_table_dictionary[_row_index][FittingKeys.row_index]
            list_row.append(_row)
        set_list_row = set(list(list_row))
        return len(set_list_row)

    def nbr_column(self):
        list_column = []
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        for _row_index in kropff_table_dictionary.keys():
            _row = kropff_table_dictionary[_row_index][FittingKeys.column_index]
            list_column.append(_row)
        set_list_column = set(list(list_column))
        return len(set_list_column)



import numpy as np
from scipy import special


def kropff_high_lambda(lda, a0, b0):
    """Equation 7.2 found in Development and application of Bragg edge neutron transmission
    imaging on the IMAT beamline. Thesis by Ranggi Sahmura Ramadhan. June 2019
    :param
    lda (Lambda - Angstroms)
    a0 parameter to fit
    b0 parameter to fit
    """
    if (a0 == np.nan) or (b0 == np.nan):
        return None
    exp_expression = np.exp(-(a0 + b0 * lda))
    return exp_expression


def kropff_low_lambda(lda, a0, b0, ahkl, bhkl):
    """Equation 7.3 found in Development and application of Bragg edge neutron transmission
    imaging on the IMAT beamline. Thesis by Ranggi Sahmura Ramadhan. June 2019
    :param
    lda (Lambda - Angstroms)
    a0 fix parameter
    b0 fix parameter
    ahkl parameter to fit
    bhkl parameter to fit
    """
    if (a0 == np.nan) or (b0 == np.nan) or (ahkl == np.nan) or (bhkl == np.nan):
        return None
    exp_expression_1 = np.exp(-(a0 + b0 * lda))
    exp_expression_2 = np.exp(-(ahkl + bhkl * lda))
    return exp_expression_1 * exp_expression_2


def kropff_bragg_peak_tof(lda, a0, b0, ahkl, bhkl, ldahkl, sigma, tau):
    """Equation 4.3 and 4.4 found in Development and application of Bragg edge neutron transmission
    imaging on the IMAT beamline. Thesis by Ranggi Sahmura Ramadhan. June 2019
    :param
    lda (Lambda - Angstroms)
    a0 fix parameter
    b0 fix parameter
    ahkl fix parameter
    bhkl fix parameter
    ldahkl parameter to fit
    tau parameter to fit
    sigma parameter to fit
    """

    def B(ldahkl, sigma, tau, lda):
        const1 = (sigma * sigma) / (2 * tau * tau)
        const2 = sigma / tau

        part1 = special.erfc(-(lda - ldahkl) / (np.sqrt(2) * sigma))
        part2 = np.exp((-(lda - ldahkl) / tau) + const1)
        part3 = special.erfc((-(lda - ldahkl) / (np.sqrt(2) * sigma)) + const2)
        return 0.5 * (part1 - part2 * part3)

    exp_expression_1 = np.exp(-(a0 + b0 * lda))
    exp_expression_2 = np.exp(-(ahkl + bhkl * lda))

    final_part1 = B(ldahkl, sigma, tau, lda) * exp_expression_1
    final_part2 = (1 - B(ldahkl, sigma, tau, lda)) * exp_expression_1 * exp_expression_2

    return final_part1 + final_part2



#!/usr/bin/env python
"""Materials model for fitting."""

from typing import Dict, List, Tuple, Union

from neutronbraggedge.braggedge import BraggEdge

from ibeatles.core.config import Material


def get_bragg_edges(
    material_config: Material,
) -> Dict[str, Union[str, float, List[float]]]:
    """
    Get Bragg edge information based on material configuration.

    Parameters
    ----------
    material_config : Material
        Material configuration from IBeatlesUserConfig

    Returns
    -------
    Dict[str, Union[str, float, List[float]]]
        Dictionary containing:
        - 'name': Material name
        - 'crystal_structure': Crystal structure
        - 'lattice': Lattice parameter
        - 'bragg_edges': List of Bragg edge values

    Examples
    --------
    >>> material_config = Material(element='Fe')
    >>> edges = get_bragg_edges(material_config)
    >>> print(edges['bragg_edges'][0])  # First Bragg edge
    4.0537

    >>> material_config = Material(
    ...     custom_material=CustomMaterial(
    ...         name='Custom',
    ...         lattice=3.52,
    ...         crystal_structure='BCC',
    ...         hkl_lambda_pairs={(1,1,0): 2.8664, (2,0,0): 2.0267}
    ...     )
    ... )
    >>> edges = get_bragg_edges(material_config)
    """
    if material_config.element is not None:
        # Use predefined element
        # TODO: the number of edges should be computed based on the maximum HKL range
        try:
            material = BraggEdge(
                material=material_config.element,
                number_of_bragg_edges=10,
            )
            return {
                "name": material_config.element,
                "crystal_structure": material.metadata["crystal_structure"],
                "lattice": material.metadata["lattice"].get(material_config.element),
                "bragg_edges": material.bragg_edges[material_config.element],
            }
        except Exception as e:
            raise ValueError(f"Error getting Bragg edges for {material_config.element}: {str(e)}")

    elif material_config.custom_material is not None:
        # Use custom material
        custom = material_config.custom_material

        # Create material definition for neutronbraggedge
        new_material = [
            {
                "name": custom.name,
                "lattice": custom.lattice,
                "crystal_structure": custom.crystal_structure,
            }
        ]

        try:
            material = BraggEdge(
                material=custom.name,
                new_material=new_material,
                number_of_bragg_edges=len(custom.hkl_lambda_pairs),
            )

            return {
                "name": custom.name,
                "crystal_structure": custom.crystal_structure,
                "lattice": custom.lattice,
                "bragg_edges": material.bragg_edges[custom.name],
            }

        except Exception as e:
            raise ValueError(f"Error creating custom material {custom.name}: {str(e)}")

    else:
        raise ValueError("No material specified in configuration")


def get_initial_bragg_edge_lambda(material_config: Material, lambda_range: Tuple[float, float]) -> float:
    """
    Get initial guess for Bragg edge wavelength within specified range.

    Parameters
    ----------
    material_config : Material
        Material configuration from IBeatlesUserConfig
    lambda_range : Tuple[float, float]
        (min, max) wavelength range to consider

    Returns
    -------
    float
        Initial guess for Bragg edge wavelength
    """
    edges = get_bragg_edges(material_config)
    bragg_edges = edges["bragg_edges"]

    # Find edges within range
    valid_edges = [edge for edge in bragg_edges if lambda_range[0] <= edge <= lambda_range[1]]

    if not valid_edges:
        raise ValueError(f"No Bragg edges found in range {lambda_range}")

    # Use middle edge in range as initial guess
    return valid_edges[len(valid_edges) // 2]



#!/usr/bin/env python
"""
Display of the d 2D image or the strain mapping 2D image
"""

import numpy as np
from matplotlib.image import _resample
from matplotlib.transforms import Affine2D

from ibeatles.step6 import ParametersToDisplay
from ibeatles.step6.get import Get


class Display:
    histo_widget = None
    image_width = None
    image_height = None

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

        self.image_height = self.parent.image_size["height"]
        self.image_width = self.parent.image_size["width"]

        o_get = Get(parent=self.parent)
        self.parameters_to_display = o_get.parameter_to_display()

    def run(self):
        o_get = Get(parent=self.parent)
        if o_get.parameter_to_display() == ParametersToDisplay.d:
            self.d_array()
        else:
            self.strain_mapping()
        self.parent.ui.stackedWidget.setCurrentIndex(1)  # 0 for debugging, 1 otherwise
        self.cleanup()

    def cleanup(self):
        self.parent.previous_parameters_displayed = self.parameters_to_display

    def d_array(self):
        """display of the d 2d image overlapping the sample image"""
        d_array = self.parent.compact_d_array
        max_value = np.nanmax(d_array)
        d_array = d_array / max_value

        data_array = d_array
        post_correction_coefficient = max_value
        parameter_to_display = ParametersToDisplay.d

        self.display_array(
            data_array=data_array,
            post_correction_coefficient=post_correction_coefficient,
            parameter_to_display=parameter_to_display,
        )

    def strain_mapping(self):
        """display of the strain mapping 2D image overlapping the sample image"""
        o_get = Get(parent=self.parent)
        strain_mapping = o_get.compact_strain_mapping()

        data_array = strain_mapping
        post_correction_coefficient = 1
        parameter_to_display = ParametersToDisplay.strain_mapping

        self.display_array(
            data_array=data_array,
            post_correction_coefficient=post_correction_coefficient,
            parameter_to_display=parameter_to_display,
        )

    def display_array(
        self,
        data_array=None,
        post_correction_coefficient=1,
        parameter_to_display=ParametersToDisplay.d,
    ):
        o_get = Get(parent=self.parent)
        integrated_image = o_get.integrated_image()
        interpolation_method = o_get.interpolation_method()
        cmap = o_get.cmap()
        scale_factor = self.parent.bin_size
        out_dimensions = (
            data_array.shape[0] * scale_factor,
            data_array.shape[1] * scale_factor,
        )
        transform = Affine2D().scale(scale_factor, scale_factor)

        self.parent.ui.matplotlib_interpolation_plot.axes.cla()
        img = self.parent.ui.matplotlib_interpolation_plot.axes.imshow(
            data_array, cmap=cmap, interpolation=interpolation_method
        )
        self.parent.ui.matplotlib_interpolation_plot.draw()

        interpolated = _resample(img, data_array, out_dimensions, transform=transform)

        self.parent.ui.matplotlib_interpolation_plot.axes.cla()
        self.parent.ui.matplotlib_interpolation_plot.axes.imshow(interpolated, cmap=cmap)
        self.parent.ui.matplotlib_interpolation_plot.draw()

        interpolated *= post_correction_coefficient

        # with overlap
        interpolated_d_array_2d = np.empty((self.image_height, self.image_width))
        interpolated_d_array_2d[:] = np.nan

        [y0, x0] = self.parent.top_left_corner_of_roi
        inter_height, inter_width = np.shape(interpolated)
        interpolated_d_array_2d[y0 : y0 + inter_height, x0 : x0 + inter_width] = interpolated

        self.parent.ui.matplotlib_plot.axes.cla()
        self.parent.ui.matplotlib_plot.axes.imshow(integrated_image, cmap="gray", vmin=0, vmax=1)
        self.parent.ui.matplotlib_plot.draw()

        min_value = self.parent.min_max[parameter_to_display]["min"]
        max_value = self.parent.min_max[parameter_to_display]["max"]

        im = self.parent.ui.matplotlib_plot.axes.imshow(
            interpolated_d_array_2d,
            interpolation=interpolation_method,
            vmin=min_value,
            vmax=max_value,
            cmap=cmap,
            alpha=0.5,
        )
        if self.parent.colorbar:
            self.parent.colorbar.mappable = im
            self.parent.colorbar.update_normal(im)
            # self.parent.colorbar.remove()
        else:
            self.parent.colorbar = self.parent.ui.matplotlib_plot.fig.colorbar(
                im, ax=self.parent.ui.matplotlib_plot.axes
            )

        self.parent.ui.matplotlib_plot.draw()

        #
        #
        # integrated_image = o_get.integrated_image()
        # interpolation_method = o_get.interpolation_method()
        # cmap = o_get.cmap()
        # scale_factor = self.parent.bin_size
        # out_dimensions = (strain_mapping.shape[0] * scale_factor,
        #                   strain_mapping.shape[1] * scale_factor)
        # transform = Affine2D().scale(scale_factor, scale_factor)
        #
        # self.parent.ui.matplotlib_interpolation_plot.axes.cla()
        # img = self.parent.ui.matplotlib_interpolation_plot.axes.imshow(strain_mapping,
        #                                                                cmap=cmap,
        #                                                                interpolation=interpolation_method)
        # self.parent.ui.matplotlib_interpolation_plot.draw()
        #
        # interpolated = _resample(img, strain_mapping, out_dimensions, transform=transform)
        #
        # self.parent.ui.matplotlib_interpolation_plot.axes.cla()
        # self.parent.ui.matplotlib_interpolation_plot.axes.imshow(interpolated, cmap=cmap)
        # self.parent.ui.matplotlib_interpolation_plot.draw()
        #
        # # with overlap
        # interpolated_strain_mapping_2d = np.empty((self.image_height, self.image_width))
        # interpolated_strain_mapping_2d[:] = np.nan
        #
        # [y0, x0] = self.parent.top_left_corner_of_roi
        # inter_height, inter_width = np.shape(interpolated)
        # interpolated_strain_mapping_2d[y0: y0+inter_height, x0: x0+inter_width] = interpolated
        #
        # self.parent.ui.matplotlib_plot.axes.cla()
        # self.parent.ui.matplotlib_plot.axes.imshow(integrated_image, cmap='gray', vmin=0, vmax=1)
        # self.parent.ui.matplotlib_plot.draw()
        #
        # min_value = self.parent.min_max[ParametersToDisplay.strain_mapping]['min']
        # max_value = self.parent.min_max[ParametersToDisplay.strain_mapping]['max']
        #
        # im = self.parent.ui.matplotlib_plot.axes.imshow(interpolated_strain_mapping_2d,
        #                                                 interpolation=interpolation_method,
        #                                                 vmin=min_value,
        #                                                 vmax=max_value,
        #                                                 cmap=cmap,
        #                                                 alpha=0.5)
        #
        # if self.parent.colorbar:
        #     self.parent.colorbar.remove()
        #
        # self.parent.colorbar = self.parent.ui.matplotlib_plot.fig.colorbar(im,
        #                                                                    ax=self.parent.ui.matplotlib_plot.axes)
        # self.parent.ui.matplotlib_plot.draw()



#!/usr/bin/env python
"""
Load load data tab
"""

import os

from ibeatles import DataType
from ibeatles.all_steps.material import Material as AllStepsMaterial
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.step1.data_handler import DataHandler
from ibeatles.step1.gui_handler import Step1GuiHandler
from ibeatles.step2.plot import Step2Plot
from ibeatles.utilities.gui_handler import GuiHandler
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class LoadLoadDataTab:
    def __init__(self, parent=None):
        self.parent = parent
        self.session_dict = parent.session_dict
        self.combine_algo = "sum" if self.parent.ui.roi_add_button.isChecked() else "mean"

    def sample(self):
        session_dict = self.session_dict
        data_type = DataType.sample

        list_sample_files = self.session_dict[data_type][SessionSubKeys.list_files]
        if list_sample_files:
            input_folder = session_dict[data_type][SessionSubKeys.current_folder]
            self.parent.image_view_settings[data_type]["state"] = session_dict[data_type][
                SessionSubKeys.image_view_state
            ]
            o_data_handler = DataHandler(parent=self.parent, data_type=data_type)
            list_sample_files_fullname = [os.path.join(input_folder, _file) for _file in list_sample_files]
            o_data_handler.load_files(list_of_files=list_sample_files_fullname)
            time_spectra_file = session_dict[data_type][SessionSubKeys.time_spectra_filename]
            o_data_handler.load_time_spectra(time_spectra_file=time_spectra_file)
            list_files_selected = session_dict[data_type][SessionSubKeys.list_files_selected]
            if not list_files_selected:
                list_files_selected = [0]
            self.parent.list_roi[data_type] = session_dict[data_type][SessionSubKeys.list_rois]
            o_gui = Step1GuiHandler(parent=self.parent, data_type=data_type)
            o_gui.initialize_rois_and_labels()
            self.parent.ui.list_sample.blockSignals(True)
            for _row_selected in list_files_selected:
                _item = self.parent.ui.list_sample.item(_row_selected)
                _item.setSelected(True)
            self.parent.ui.list_sample.blockSignals(False)
            o_gui.check_time_spectra_widgets()
            o_gui.check_step1_widgets()
            # self.parent.check_files_error()
            self.parent.retrieve_general_infos(data_type=data_type)
            self.parent.retrieve_general_data_infos(data_type=data_type)
            self.parent.infos_window_update(data_type=data_type)

            o_step2_plot = Step2Plot(parent=self.parent)
            o_step2_plot.prepare_data()
            o_step2_plot.init_roi_table()

            o_pyqt = PyqtgraphUtilities(
                parent=self.parent,
                image_view=self.parent.ui.image_view,
                data_type=data_type,
            )
            o_pyqt.set_state(session_dict[data_type][SessionSubKeys.image_view_state])
            histogram_level = session_dict[data_type][SessionSubKeys.image_view_histogram][self.combine_algo]
            o_pyqt.set_histogram_level(histogram_level=histogram_level)

    def ob(self):
        session_dict = self.session_dict
        data_type = DataType.ob

        self.parent.image_view_settings[data_type]["state"] = session_dict[data_type][SessionSubKeys.image_view_state]
        # self.parent.image_view_settings[data_type]['histogram'][self.combine_algo] = \
        #     session_dict[data_type][SessionSubKeys.image_view_histogram]
        list_ob_files = session_dict[data_type][SessionSubKeys.list_files]
        if list_ob_files:
            input_folder = session_dict[data_type][SessionSubKeys.current_folder]
            o_data_handler = DataHandler(parent=self.parent, data_type=data_type)
            list_ob_files_fullname = [os.path.join(input_folder, _file) for _file in list_ob_files]
            o_data_handler.load_files(list_of_files=list_ob_files_fullname)
        list_files_selected = session_dict[data_type][SessionSubKeys.list_files_selected]
        self.parent.list_roi[data_type] = session_dict[data_type][SessionSubKeys.list_rois]
        o_gui = Step1GuiHandler(parent=self.parent, data_type=data_type)
        o_gui.initialize_rois_and_labels()
        self.parent.ui.list_open_beam.blockSignals(True)
        for _row_selected in list_files_selected:
            _item = self.parent.ui.list_open_beam.item(_row_selected)
            _item.setSelected(True)
        self.parent.ui.list_open_beam.blockSignals(False)

        o_pyqt = PyqtgraphUtilities(
            parent=self.parent,
            image_view=self.parent.ui.ob_image_view,
            data_type=data_type,
        )
        o_pyqt.set_state(session_dict[data_type][SessionSubKeys.image_view_state])
        o_pyqt.reload_histogram_level()
        histogram_level = session_dict[data_type][SessionSubKeys.image_view_histogram][self.combine_algo]
        o_pyqt.set_histogram_level(histogram_level=histogram_level)

    def instrument(self):
        session_dict = self.session_dict

        o_gui = GuiHandler(parent=self.parent)
        list_ui_data = {
            "distance": self.parent.ui.distance_source_detector,
            "beam": self.parent.ui.beam_rate,
            "detector": self.parent.ui.detector_offset,
        }

        for _key in list_ui_data.keys():
            list_ui_data[_key].blockSignals(True)

        o_gui.set_index_selected(
            index=session_dict[SessionKeys.instrument][SessionSubKeys.beam_index],
            ui=list_ui_data["beam"],
        )
        o_gui.set_text(
            value=session_dict[SessionKeys.instrument][SessionSubKeys.distance_source_detector],
            ui=list_ui_data["distance"],
        )

        for _key in list_ui_data.keys():
            list_ui_data[_key].blockSignals(False)

        o_gui.set_text(
            value=session_dict[SessionKeys.instrument][SessionSubKeys.detector_value],
            ui=list_ui_data["detector"],
        )

    def material(self):
        material_session_dict = self.session_dict[SessionKeys.material]

        material_mode = material_session_dict[SessionSubKeys.material_mode]
        o_gui = GuiHandler(parent=self.parent)
        o_gui.set_material_active_tab(active_tab_mode=material_mode)

        # pre-defined
        pre_defined_material_index = material_session_dict[SessionSubKeys.pre_defined][
            SessionSubKeys.pre_defined_selected_element_index
        ]
        self.parent.ui.pre_defined_list_of_elements.setCurrentIndex(pre_defined_material_index)

        # custom
        custom_material_name = material_session_dict[SessionSubKeys.custom_material_name]
        self.parent.ui.user_defined_element_name.setText(custom_material_name)

        # method 1
        index_of_element = material_session_dict[SessionSubKeys.custom_method1][
            SessionSubKeys.user_defined_fill_fields_with_element_index
        ]
        self.parent.ui.user_defined_list_of_elements.blockSignals(True)
        self.parent.ui.user_defined_list_of_elements.setCurrentIndex(index_of_element)
        self.parent.ui.user_defined_list_of_elements.blockSignals(False)

        lattice = material_session_dict[SessionSubKeys.custom_method1][SessionSubKeys.lattice]
        if not (lattice == ""):
            self.parent.ui.method1_lattice_value_2.blockSignals(True)
            self.parent.ui.method1_lattice_value_2.setText(f"{float(lattice):.3f}")
            self.parent.ui.method1_lattice_value_2.blockSignals(False)

        crystal_structure_index = material_session_dict[SessionSubKeys.custom_method1][
            SessionSubKeys.crystal_structure_index
        ]
        self.parent.ui.method1_crystal_value_2.blockSignals(True)
        self.parent.ui.method1_crystal_value_2.setCurrentIndex(crystal_structure_index)
        self.parent.ui.method1_crystal_value_2.blockSignals(False)

        method1_table = material_session_dict[SessionSubKeys.custom_method1][SessionSubKeys.material_hkl_table]
        method1_column_names = material_session_dict[SessionSubKeys.custom_method1][SessionSubKeys.column_names]
        self.parent.ui.method1_tableWidget.blockSignals(True)
        GuiHandler.fill_table_data(
            table_ui=self.parent.ui.method1_tableWidget,
            table_dict=method1_table,
            column_names=method1_column_names,
        )
        self.parent.ui.method1_tableWidget.blockSignals(False)

        method2_table = material_session_dict[SessionSubKeys.custom_method2][SessionSubKeys.material_hkl_table]
        method2_column_names = material_session_dict[SessionSubKeys.custom_method2][SessionSubKeys.column_names]
        GuiHandler.fill_table_data(
            table_ui=self.parent.ui.method2_tableWidget,
            table_dict=method2_table,
            column_names=method2_column_names,
        )

        o_material = AllStepsMaterial(parent=self.parent)
        o_material.check_status_of_all_widgets()



#!/usr/bin/env python
"""
GUI handler for the step2
"""

import numpy as np
import pyqtgraph as pg
from qtpy.QtWidgets import QCheckBox, QComboBox, QTableWidgetItem

from ibeatles import DEFAULT_ROI, RegionType
from ibeatles.step2 import gui_handler
from ibeatles.step2.get import Get as Step2Get


class Step2RoiHandler:
    def __init__(self, parent=None):
        self.parent = parent

    def save_table(self):
        list_roi = self.parent.list_roi["normalization"]

        o_get = Step2Get(parent=self.parent)
        for _row, roi in enumerate(list_roi):
            try:
                _row_infos = o_get.roi_table_row(_row)
            except ValueError:
                return

            list_roi[_row] = _row_infos

        self.parent.list_roi["normalization"] = list_roi

    def enable_selected_roi(self):
        list_roi = self.parent.list_roi["normalization"]
        list_roi_id = self.parent.list_roi_id["normalization"]
        list_label_roi_id = self.parent.list_label_roi_id["normalization"]

        for index, roi in enumerate(list_roi):
            _roi_id = list_roi_id[index]
            _label_roi_id = list_label_roi_id[index]
            is_roi_visible = roi[0]
            _roi_id.setVisible(is_roi_visible)
            _label_roi_id.setVisible(is_roi_visible)

    def save_roi(self):
        list_roi_id = self.parent.list_roi_id["normalization"]
        list_roi = self.parent.list_roi["normalization"]

        sample = self.parent.data_metadata["normalization"]["data"]
        image_item = self.parent.step2_ui["image_view"].imageItem

        for _index, _roi_id in enumerate(list_roi_id):
            region = _roi_id.getArraySlice(sample, image_item)
            x0 = region[0][0].start
            x1 = region[0][0].stop - 1
            y0 = region[0][1].start
            y1 = region[0][1].stop - 1

            width = x1 - x0
            height = y1 - y0

            _roi = list_roi[_index]
            _roi[1] = x0
            _roi[2] = y0
            _roi[3] = width
            _roi[4] = height

            list_roi[_index] = _roi

        self.parent.list_roi["normalization"] = list_roi

    def remove_roi(self):
        selection = self.parent.ui.normalization_tableWidget.selectedRanges()
        if len(selection) == 0:
            return

        selection = selection[0]
        _row_selected = selection.bottomRow()

        self.parent.ui.normalization_tableWidget.removeRow(_row_selected)

        list_roi = self.parent.list_roi["normalization"]
        list_roi_id = self.parent.list_roi_id["normalization"]
        list_label_roi_id = self.parent.list_label_roi_id["normalization"]

        new_list_roi = []
        new_list_roi_id = []
        new_list_label_roi_id = []
        for _index, _roi in enumerate(list_roi):
            if _index == _row_selected:
                self.parent.step2_ui["image_view"].removeItem(list_roi_id[_index])
                self.parent.step2_ui["image_view"].removeItem(list_label_roi_id[_index])
                continue
            new_list_roi.append(_roi)
            new_list_roi_id.append(list_roi_id[_index])
            new_list_label_roi_id.append(list_label_roi_id[_index])

        self.parent.list_roi["normalization"] = new_list_roi
        self.parent.list_roi_id["normalization"] = new_list_roi_id
        self.parent.list_label_roi_id["normalization"] = new_list_label_roi_id

        o_gui = gui_handler.Step2GuiHandler(parent=self.parent)
        o_gui.check_add_remove_roi_buttons()

    def add_roi_in_image(self):
        x0 = int(DEFAULT_ROI[1])
        y0 = int(DEFAULT_ROI[2])
        width = int(DEFAULT_ROI[3])
        height = int(DEFAULT_ROI[4])

        roi = pg.ROI([x0, y0], [width, height])
        roi.addScaleHandle([1, 1], [0, 0])
        roi.sigRegionChanged.connect(self.parent.normalization_manual_roi_changed)
        self.parent.step2_ui["image_view"].addItem(roi)

        label_roi = pg.TextItem(
            html='<div style="text-align: center"><span style="color: '
            '#ff0000;">' + RegionType.background + "</span></div>",
            anchor=(-0.3, 1.3),
            border="w",
            fill=(0, 0, 255, 50),
        )
        label_roi.setPos(x0, y0)
        self.parent.step2_ui["image_view"].addItem(label_roi)

        return roi, label_roi

    def add_roi(self):
        nbr_row_table = self.parent.ui.normalization_tableWidget.rowCount()
        new_roi_id, new_label_roi_id = self.add_roi_in_image()

        self.parent.list_roi["normalization"].append(self.parent.init_array_normalization)
        self.parent.list_roi_id["normalization"].append(new_roi_id)
        self.parent.list_label_roi_id["normalization"].append(new_label_roi_id)
        self.insert_row(row=nbr_row_table)

        o_gui = gui_handler.Step2GuiHandler(parent=self.parent)
        o_gui.check_add_remove_roi_buttons()

    def get_item(self, text):
        _item = QTableWidgetItem(text)
        # _item.setBackground(color)
        return _item

    def insert_row(self, row=-1):
        self.parent.ui.normalization_tableWidget.insertRow(row)

        init_array = self.parent.list_roi["normalization"][-1]
        [flag, x0, y0, width, height, region_type] = init_array

        # button
        _widget = QCheckBox()
        _widget.setChecked(flag)
        # QtCore.QObject.connect(
        #   _widget,
        #   QtCore.SIGNAL("stateChanged(int)"),
        #   self.parent.normalization_row_status_changed,
        # )
        _widget.stateChanged.connect(self.parent.normalization_row_status_changed)
        self.parent.ui.normalization_tableWidget.setCellWidget(row, 0, _widget)

        # x0
        _item = self.get_item(str(x0))
        self.parent.ui.normalization_tableWidget.setItem(row, 1, _item)

        # y0
        _item = self.get_item(str(y0))
        self.parent.ui.normalization_tableWidget.setItem(row, 2, _item)

        # width
        _item = self.get_item(str(width))
        self.parent.ui.normalization_tableWidget.setItem(row, 3, _item)

        # height
        _item = self.get_item(str(height))
        self.parent.ui.normalization_tableWidget.setItem(row, 4, _item)

        # region type
        _widget = QComboBox()
        _widget.addItems([RegionType.sample, RegionType.background])
        index = 0 if (region_type == RegionType.sample) else 1
        _widget.setCurrentIndex(index)
        _widget.currentIndexChanged.connect(self.parent.normalization_row_status_region_type_changed)
        self.parent.ui.normalization_tableWidget.setCellWidget(row, 5, _widget)

    def get_list_of_roi_to_use(self):
        list_roi = []

        nbr_row = self.parent.ui.normalization_tableWidget.rowCount()
        o_get = Step2Get(parent=self.parent)
        for _row in np.arange(nbr_row):
            _row_value = o_get.roi_table_row(row=_row)
            if _row_value[0]:
                _roi = _row_value[1:5]
                list_roi.append(_roi)

        return list_roi

    def get_list_of_background_roi_to_use(self):
        list_background_roi = []

        nbr_row = self.parent.ui.normalization_tableWidget.rowCount()
        o_get = Step2Get(parent=self.parent)
        for _row in np.arange(nbr_row):
            _row_value = o_get.roi_table_row(row=_row)
            if _row_value[0] and (_row_value[-1] == RegionType.background):
                _roi = _row_value[1:5]
                list_background_roi.append(_roi)

        return list_background_roi



#!/usr/bin/env python
"""Data loading functions."""

import glob
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Tuple

import numpy as np
from astropy.io import fits
from loguru import logger
from neutronbraggedge.experiment_handler.experiment import Experiment
from neutronbraggedge.experiment_handler.tof import TOF
from PIL import Image
from tqdm.auto import tqdm


def cleanup_list_of_files(list_of_files: List[str], base_number: int = 5) -> List[str]:
    """
    Keep only the files that have the same number of characters as the first n files.

    Parameters
    ----------
    list_of_files : List[str]
        List of file names to clean up.
    base_number : int, optional
        Number of files to use as a base for comparison (default is 5).

    Returns
    -------
    List[str]
        Cleaned list of file names.

    Raises
    ------
    ValueError
        If the format of the input files does not match.
    """
    if len(list_of_files) == 0:
        return []

    len_base_files = [len(file) for file in list_of_files[:base_number]]

    # Make sure all the lengths of the base number files match
    if len(set(len_base_files)) > 1:
        raise ValueError("Format Input File Do Not Match!")

    len_file = len_base_files[0]
    return [file for file in list_of_files if len(file) == len_file]


def load_image(file_path: str) -> Tuple[np.ndarray, Dict[str, Any]]:
    """
    Load an image file and return its data and metadata.

    Parameters
    ----------
    file_path : str
        Path to the image file.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, Any]]
        A tuple containing the image data as a numpy array and a dictionary of metadata.
    """
    file_extension = os.path.splitext(file_path)[1].lower()
    if file_extension in [".tiff", ".tif"]:
        return load_tiff(file_path)
    elif file_extension == ".fits":
        return load_fits(file_path)
    else:
        raise ValueError(f"Unsupported file format: {file_extension}")


def load_tiff(file_path: str) -> Tuple[np.ndarray, Dict[str, Any]]:
    """
    Load a TIFF image file.

    Parameters
    ----------
    file_path : str
        Path to the TIFF file.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, Any]]
        A tuple containing the image data as a numpy array and a dictionary of metadata.
    """
    with Image.open(file_path) as img:
        data = np.array(img)
        # Replace NaN and Inf values with 0 to avoid computation issues
        data = np.nan_to_num(data, nan=0, posinf=0, neginf=0)
        metadata = img.tag_v2 if hasattr(img, "tag_v2") else {}

    return data, process_tiff_metadata(metadata, data, file_path)


def load_fits(file_path: str) -> Tuple[np.ndarray, Dict[str, Any]]:
    """
    Load a FITS image file.

    Parameters
    ----------
    file_path : str
        Path to the FITS file.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, Any]]
        A tuple containing the image data as a numpy array and a dictionary of metadata.
    """
    with fits.open(file_path) as hdul:
        data = hdul[0].data
        # Replace NaN and Inf values with 0 to avoid computation issues
        data = np.nan_to_num(data, nan=0, posinf=0, neginf=0)
        header = hdul[0].header

    return data, process_fits_metadata(header, data, file_path)


def process_tiff_metadata(metadata: Dict[str, Any], data: np.ndarray, file_path: str) -> Dict[str, Dict[str, Any]]:
    """
    Process TIFF metadata and extract relevant information.

    Parameters
    ----------
    metadata : Dict[str, Any]
        Raw metadata dictionary from the TIFF file.
    data : np.ndarray
        Image data.
    file_path : str
        Path to the TIFF file.

    Returns
    -------
    Dict[str, Dict[str, Any]]
        Processed metadata dictionary.
    """
    processed_metadata = {
        "acquisition_duration": {"name": "Acquisition Duration", "value": 0},
        "acquisition_time": {"name": "Acquisition Time", "value": ""},
        "image_size": {"name": "Image(s) Size", "value": ""},
        "image_type": {"name": "Image Type", "value": ""},
        "min_counts": {"name": "min counts", "value": 0},
        "max_counts": {"name": "max counts", "value": 0},
    }

    # Acquisition time
    acquisition_time = metadata.get(65000, [metadata.get(279, [None])])[0]
    if acquisition_time is None:
        acquisition_time = os.path.getmtime(file_path)
    processed_metadata["acquisition_time"]["value"] = acquisition_time

    # Acquisition duration
    acquisition_duration = metadata.get(65021, ["N/A"])[0]
    if isinstance(acquisition_duration, str) and ":" in acquisition_duration:
        acquisition_duration = acquisition_duration.split(":")[1]
    processed_metadata["acquisition_duration"]["value"] = acquisition_duration

    # Image size
    size_x = metadata.get(65028, [metadata.get(256, [data.shape[1]])])[0]
    size_y = metadata.get(65029, [metadata.get(257, [data.shape[0]])])[0]
    if isinstance(size_x, str) and ":" in size_x:
        size_x = size_x.split(":")[1]
    if isinstance(size_y, str) and ":" in size_y:
        size_y = size_y.split(":")[1]
    processed_metadata["image_size"]["value"] = f"{size_x} x {size_y}"

    # Image type
    bits = metadata.get(258, [data.dtype.itemsize * 8])[0]
    processed_metadata["image_type"]["value"] = f"{bits} bits"

    # Min and max counts
    processed_metadata["min_counts"]["value"] = np.min(data)
    processed_metadata["max_counts"]["value"] = np.max(data)

    return processed_metadata


def process_fits_metadata(header: fits.Header, data: np.ndarray, file_path: str) -> Dict[str, Dict[str, Any]]:
    """
    Process FITS metadata and extract relevant information.

    Parameters
    ----------
    header : fits.Header
        FITS file header.
    data : np.ndarray
        Image data.
    file_path : str
        Path to the FITS file.

    Returns
    -------
    Dict[str, Dict[str, Any]]
        Processed metadata dictionary.
    """
    processed_metadata = {
        "acquisition_duration": {"name": "Acquisition Duration", "value": 0},
        "acquisition_time": {"name": "Acquisition Time", "value": ""},
        "image_size": {"name": "Image(s) Size", "value": ""},
        "image_type": {"name": "Image Type", "value": ""},
        "min_counts": {"name": "min counts", "value": 0},
        "max_counts": {"name": "max counts", "value": 0},
    }

    processed_metadata["acquisition_time"]["value"] = header.get("DATE", os.path.getmtime(file_path))
    processed_metadata["acquisition_duration"]["value"] = header.get("EXPOSURE", header.get("TIMEBIN", "N/A"))
    processed_metadata["image_size"]["value"] = (
        f"{header.get('NAXIS1', data.shape[1])} x {header.get('NAXIS2', data.shape[0])}"
    )
    processed_metadata["image_type"]["value"] = f"{header.get('BITPIX', data.dtype.itemsize * 8)} bits"
    processed_metadata["min_counts"]["value"] = np.min(data)
    processed_metadata["max_counts"]["value"] = np.max(data)

    return processed_metadata


def load_data_from_folder(folder: str, file_extension: str = ".tif") -> Dict[str, Any]:
    """
    Load raw data from a specified folder.

    Parameters
    ----------
    folder : str
        Path to the folder containing raw data files.
    file_extension : str, optional
        File extension of the raw data files (default is '.tif').

    Returns
    -------
    Dict[str, Any]
        Dictionary containing loaded raw data and metadata.
    """
    file_list = glob.glob(os.path.join(folder, f"*{file_extension}"))
    file_list.sort()

    # Clean up the list of files
    short_list_of_files = [os.path.basename(file) for file in file_list]
    short_list_of_files = cleanup_list_of_files(short_list_of_files)

    data = []
    metadata = []
    for file in tqdm(short_list_of_files):
        full_file_name = os.path.join(folder, file)
        img_data, img_metadata = load_image(full_file_name)
        data.append(img_data)
        metadata.append(img_metadata)

    # Get image size from the first image
    if data:
        height, width = data[0].shape
        size = {"width": width, "height": height}
    else:
        size = {"width": 0, "height": 0}

    logger.info(f"Loaded {len(data)} images from folder: {folder}")

    return {
        "data": np.array(data),
        "metadata": metadata,
        "file_list": short_list_of_files,
        "folder": folder,
        "size": size,
    }


def get_filename_with_pattern(folder: str, match_pattern: str) -> str:
    logging.info(f"Searching for file in folder: {folder} with pattern: {match_pattern}")
    search_path = Path(folder) / match_pattern
    str_search_path = str(search_path)
    str_search_path = str_search_path.strip()
    logging.info(f"\tSearch path: {search_path}")
    search_results = sorted(glob.glob(str_search_path))
    logging.info(f"\tFound {len(search_results)} matching files.")

    if search_results and os.path.exists(search_results[0]):
        logging.info(f"\tUsing file: {search_results[0]}")
        return search_results[0]
    else:
        return ""


def get_time_spectra_filename(folder: str, match_pattern: str = "*_Spectra.txt") -> str:
    """
    Find the time spectra file in the folder.

    Parameters
    ----------
    folder : str
        Folder to search for the time spectra file.
    match_pattern : str, optional
        Pattern to match the time spectra file, unix style (default is "*_Spectra.txt").

    Returns
    -------
    str
        Full path to the time spectra file, or an empty string if not found.
    """
    logging.info(f"Searching for time spectra file in folder: {folder} with pattern: {match_pattern}")
    return get_filename_with_pattern(folder, match_pattern)


def get_shutter_count_filename(folder: str, match_pattern: str = "*_ShutterCount.txt") -> str:
    """
    Find the shutter count file in the folder.

    Parameters
    ----------
    folder : str
        Folder to search for the shutter count file.
    match_pattern : str, optional
        Pattern to match the shutter count file, unix style (default is "*_ShutterCount.txt").

    Returns
    -------
    str
        Full path to the shutter count file, or an empty string if not found.
    """
    logging.info(f"Searching for shutter count file in folder: {folder} with pattern: {match_pattern}")
    return get_filename_with_pattern(folder, match_pattern)


def load_time_spectra(
    file_path: str,
    distance_source_detector_m: float,
    detector_offset_micros: float,
) -> Dict[str, Any]:
    """
    Load time spectra data from a specified file and calculate lambda scale.

    Parameters
    ----------
    file_path : str
        Path to the time spectra file.
    distance_source_detector_m : float
        Distance from the source to the detector in meters.
    detector_offset_micros : float
        Detector offset in microseconds

    Returns
    -------
    Dict[str, Any]
        Dictionary containing loaded time spectra data and calculated lambda array.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Time spectra file not found: {file_path}")

    tof_handler = TOF(filename=file_path)
    tof_array = tof_handler.tof_array
    counts_array = tof_handler.counts_array

    # Calculate lambda scale
    exp = Experiment(
        tof=tof_array,
        distance_source_detector_m=distance_source_detector_m,
        detector_offset_micros=detector_offset_micros,
    )
    lambda_array = exp.lambda_array

    logger.info(f"Loaded time spectra data from file: {file_path}")

    return {
        "filename": file_path,
        "short_filename": os.path.basename(file_path),
        "tof_array": tof_array,
        "counts_array": counts_array,
        "lambda_array": lambda_array,
    }



#!/usr/bin/env python
"""
This module provides a class to handle GUI.
"""

import numpy as np

from ibeatles import DataType, XAxisMode
from ibeatles.session import MaterialMode
from ibeatles.utilities.table_handler import TableHandler


class GuiHandler:
    def __init__(self, parent=None):
        self.parent = parent

    def get_active_tab(self):
        """return either 'sample', 'ob', 'normalization' or 'normalized'"""
        top_tab_index = self.parent.ui.tabWidget.currentIndex()

        if top_tab_index == 1:
            return DataType.normalization
        if top_tab_index == 2:
            return DataType.normalized
        if top_tab_index == 0:
            load_data_tab_index = self.parent.ui.load_data_tab.currentIndex()
            if load_data_tab_index == 0:
                return DataType.sample
            if load_data_tab_index == 1:
                return DataType.ob

    def set_tab(self, tab_index=0):
        self.parent.ui.tabWidget.setCurrentIndex(tab_index)

    def get_material_active_tab(self):
        """
        return either MaterialMode.pre_defined, MaterialMode.custom_method1 or MaterialMode.custom_method2
        """
        top_tab_index = self.parent.ui.material_top_tabWidget.currentIndex()
        if top_tab_index == 0:  # pre-defined
            return MaterialMode.pre_defined

        custom_tab_index = self.parent.ui.material_custom_tabWidget.currentIndex()
        if custom_tab_index == 0:  # method 1
            return MaterialMode.custom_method1

        return MaterialMode.custom_method2

    def set_material_active_tab(self, active_tab_mode=MaterialMode.pre_defined):
        if active_tab_mode == MaterialMode.pre_defined:
            self.parent.ui.material_top_tabWidget.setCurrentIndex(0)
        else:
            self.parent.ui.material_top_tabWidget.setCurrentIndex(1)
            if active_tab_mode == MaterialMode.custom_method1:
                self.parent.ui.material_custom_tabWidget.setCurrentIndex(0)
            else:
                self.parent.ui.material_custom_tabWidget.setCurrentIndex(1)

    @staticmethod
    def collect_table_data(table_ui=None):
        """
        return as a dictionary the content of a tableWidget where the top key is the row number, and
        the under-keys are the names of the columns
        """
        if table_ui is None:
            return None

        o_table = TableHandler(table_ui=table_ui)
        column_names = o_table.get_column_names()
        nbr_column = len(column_names)
        nbr_row = o_table.row_count()

        table = {}
        for _row in np.arange(nbr_row):
            _row_entry = {}
            for _col in np.arange(nbr_column):
                _row_entry[str(column_names[_col])] = o_table.get_item_str_from_cell(row=_row, column=_col)
            table[int(_row)] = _row_entry

        return table, column_names

    @staticmethod
    def fill_table_data(table_ui=None, table_dict=None, column_names=None):
        o_table = TableHandler(table_ui=table_ui)
        o_table.remove_all_rows()
        table_ui.blockSignals(True)
        for _row in table_dict.keys():
            o_table.insert_empty_row(row=int(_row))
            for _col_key in table_dict[_row].keys():
                _col = column_names.index(_col_key)
                _val = table_dict[_row][_col_key]
                if _val is None:
                    _val = ""
                o_table.insert_item(row=int(_row), column=int(_col), value=_val)
        table_ui.blockSignals(False)

    def enable_xaxis_button(self, tof_flag=True):
        list_button_ui = self.parent.xaxis_button_ui
        active_type = self.get_active_tab()

        if tof_flag:
            for _key in list_button_ui[active_type]:
                list_button_ui[active_type][_key].setEnabled(True)
        else:
            list_button_ui[active_type]["tof"].setEnabled(False)
            list_button_ui[active_type]["lambda"].setEnabled(False)
            list_button_ui[active_type]["file_index"].setChecked(True)

    def get_xaxis_checked(self, data_type=DataType.sample):
        return self.parent.data_metadata[data_type]["xaxis"]

    def xaxis_label(self):
        o_gui = GuiHandler(parent=self.parent)
        data_type = o_gui.get_active_tab()
        button = self.get_xaxis_checked(data_type=data_type)

        if button == "file_index":
            label = "File Index"
        elif button == "tof":
            label = "TOF (\u00b5s)"
        else:
            label = "\u03bb (\u212b)"

        if data_type == "sample":
            plot_ui = self.parent.ui.bragg_edge_plot
        elif data_type == "ob":
            plot_ui = self.parent.ui.ob_bragg_edge_plot
        else:
            plot_ui = self.parent.ui.normalized_bragg_edge_plot

        plot_ui.setLabel("bottom", label)

    def get_text(self, ui=None):
        if ui is None:
            return ""
        return str(ui.text())

    def get_index_selected(self, ui=None):
        if ui is None:
            return -1
        return ui.currentIndex()

    def set_text(self, value="", ui=None):
        if ui is None:
            return
        ui.setText(value)

    def set_index_selected(self, index=-1, ui=None):
        if ui is None:
            return
        ui.setCurrentIndex(index)

    def get_text_selected(self, ui=None):
        if ui is None:
            return ""
        return str(ui.currentText())

    def get_step2_xaxis_checked(self):
        return self.parent.data_metadata[DataType.normalization]["xaxis"]

    def update_bragg_peak_scrollbar(self, xaxis_mode=XAxisMode.file_index_mode, force_hide_widgets=False):
        list_label_ui = [
            self.parent.hkl_scrollbar_ui["label"][key] for key in self.parent.hkl_scrollbar_ui["label"].keys()
        ]
        list_widget_ui = [
            self.parent.hkl_scrollbar_ui["widget"][key] for key in self.parent.hkl_scrollbar_ui["widget"].keys()
        ]

        list_ui = [*list_label_ui, *list_widget_ui]

        for _ui in list_ui:
            _ui.blockSignals(True)

        if force_hide_widgets:
            for _ui in list_ui:
                _ui.setEnabled(False)
            return

        if xaxis_mode in (XAxisMode.file_index_mode, XAxisMode.tof_mode):
            is_enable = False
        else:
            is_enable = True

        for _ui in list_ui:
            _ui.setEnabled(is_enable)

        for _ui in list_ui:
            _ui.blockSignals(False)



#!/usr/bin/env python
"""
Export for step6
"""

import logging
import os

import h5py
from NeuNorm.normalization import Normalization
from qtpy.QtWidgets import QFileDialog

from ibeatles import DataType, FileType
from ibeatles.fitting import FittingKeys
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.step6 import ParametersToDisplay
from ibeatles.step6.get import Get
from ibeatles.utilities.export import format_kropff_dict, format_kropff_table
from ibeatles.utilities.file_handler import (
    FileHandler,
    create_full_export_file_name,
)
from ibeatles.utilities.get import Get as UtilitiesGet
from ibeatles.utilities.json_handler import save_json
from ibeatles.utilities.time import get_current_time_in_special_file_name_format


class Export:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent
        self.working_dir = os.path.dirname(
            os.path.abspath((self.grand_parent.data_metadata[DataType.normalized]["folder"]))
        )

    @staticmethod
    def _make_image_base_name(normalized_folder, ext="tiff", parameters=ParametersToDisplay.d):
        base_file_name = os.path.basename(normalized_folder) + "_" + parameters + f".{ext}"
        return base_file_name

    def select_output_folder(self):
        output_folder = str(
            QFileDialog.getExistingDirectory(self.parent, "Select where to export ...", self.working_dir)
        )
        return output_folder

    def image(
        self,
        d_spacing_image=False,
        strain_mapping_image=False,
        integrated_image=False,
        output_folder=None,
    ):
        if output_folder is None:
            output_folder = str(
                QFileDialog.getExistingDirectory(self.parent, "Select where to export ...", self.working_dir)
            )

        if output_folder:
            output_folder = os.path.abspath(output_folder)
            o_get = Get(parent=self.parent)

            if d_spacing_image:
                d_spacing_file_name = Export._make_image_base_name(self.working_dir)
                full_d_output_file_name = os.path.join(output_folder, d_spacing_file_name)
                d_array = o_get.d_array()

                o_norm = Normalization()
                o_norm.load(data=d_array, notebook=False)
                o_norm.data["sample"]["file_name"][0] = d_spacing_file_name
                o_norm.export(data_type="sample", folder=output_folder)
                logging.info(f"Export d_spacing: {full_d_output_file_name}")

            if strain_mapping_image:
                strain_mapping_file_name = Export._make_image_base_name(
                    self.working_dir, parameters=ParametersToDisplay.strain_mapping
                )
                full_strain_output_file_name = os.path.join(output_folder, strain_mapping_file_name)
                strain_mapping_array = o_get.strain_mapping()

                o_norm = Normalization()
                o_norm.load(data=strain_mapping_array, notebook=False)
                o_norm.data["sample"]["file_name"][0] = strain_mapping_file_name
                o_norm.export(data_type="sample", folder=output_folder)
                logging.info(f"Export strain mapping: {full_strain_output_file_name}")

            if integrated_image:
                integrated_image_file_name = Export._make_image_base_name(
                    self.working_dir, parameters=ParametersToDisplay.integrated_image
                )
                full_image_output_file_name = os.path.join(output_folder, integrated_image_file_name)
                integrated_image = o_get.integrated_image()

                o_norm = Normalization()
                o_norm.load(data=integrated_image, notebook=False)
                o_norm.data["sample"]["file_name"][0] = integrated_image_file_name
                o_norm.export(data_type="sample", folder=output_folder)
                logging.info(f"Export strain mapping: {full_image_output_file_name}")

    def table(self, file_type=FileType.ascii, output_folder=None):
        if output_folder is None:
            output_folder = str(
                QFileDialog.getExistingDirectory(
                    self.grand_parent,
                    "Select where to export the table as an ASCII file",
                    self.working_dir,
                )
            )

        if output_folder:
            output_folder = os.path.abspath(output_folder)
            output_file_name = create_full_export_file_name(
                os.path.join(output_folder, "strain_mapping_table"), file_type
            )

            kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
            o_get = Get(parent=self.parent)
            strain_mapping_dict = o_get.strain_mapping_dictionary()

            if file_type == FileType.ascii:
                formatted_table = format_kropff_table(
                    table=kropff_table_dictionary,
                    d_dict=self.parent.d_dict,
                    strain_dict=strain_mapping_dict,
                )
                FileHandler.make_ascii_file(data=formatted_table, output_file_name=output_file_name)
            else:
                formatted_dict = format_kropff_dict(
                    table=kropff_table_dictionary,
                    d_dict=self.parent.d_dict,
                    strain_dict=strain_mapping_dict,
                )
                FileHandler.make_json_file(data_dict=formatted_dict, output_file_name=output_file_name)

            logging.info(f"Exported {file_type} strain mapping table: {output_file_name}")

    @staticmethod
    def format_kropff_table(table=None, d_dict={}, strain_dict={}):
        formatted_table = [
            "#index, "
            + "bin x0, bin y0, bin x1, bin y1, "
            + "lambda hkl val, lambda hkl err, "
            + "d value, d err, strain, strain error"
        ]
        for _row in table.keys():
            _entry = table[_row]

            _row_index = _row
            _bin_x0 = _entry["bin_coordinates"]["x0"]
            _bin_y0 = _entry["bin_coordinates"]["y0"]
            _bin_x1 = _entry["bin_coordinates"]["x1"]
            _bin_y1 = _entry["bin_coordinates"]["y1"]

            _lambda_hkl_val = _entry["lambda_hkl"]["val"]
            _lambda_hkl_err = _entry["lambda_hkl"]["err"]

            _d_value = d_dict[_row]["val"]
            _d_err = d_dict[_row]["err"]

            _strain_value = strain_dict[_row]["val"]
            _strain_value_err = strain_dict[_row]["err"]

            line = [
                _row_index,
                _bin_x0,
                _bin_y0,
                _bin_x1,
                _bin_y1,
                _lambda_hkl_val,
                _lambda_hkl_err,
                _d_value,
                _d_err,
                _strain_value,
                _strain_value_err,
            ]
            line = [str(_value) for _value in line]

            formatted_table.append(", ".join(line))

        return formatted_table

    def hdf5(self, output_folder: str = None):
        output_folder = os.path.abspath(output_folder)
        # output_file_name = os.path.join(output_folder, "strain_mapping_table.txt")
        output_file_name = create_full_export_file_name(os.path.join(output_folder, "fitting"), FileType.hdf5)

        logging.info(f"Exporting fitting table and images to hdf5 file {output_file_name}")

        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        o_get = Get(parent=self.parent)
        o_get_utilities = UtilitiesGet(parent=self.grand_parent)

        integrated_image = o_get.integrated_image()
        strain_mapping_dict = o_get.strain_mapping_dictionary()
        formatted_dict = format_kropff_dict(
            table=kropff_table_dictionary,
            d_dict=self.parent.d_dict,
            strain_dict=strain_mapping_dict,
        )

        with h5py.File(output_file_name, "w") as f:
            entry = f.create_group("entry")

            # general infos
            d0 = o_get.active_d0()
            material_name = o_get.material_name()
            hkl_value = o_get.hkl_value()
            distance_source_detector = o_get_utilities.distance_source_detector()
            detector_offset = o_get_utilities.detector_offset()
            bin_size = o_get_utilities.bin_size()

            metadata_group = entry.create_group("metadata")
            metadata_group.create_dataset("d0", data=d0)
            metadata_group.create_dataset("hkl_value", data=hkl_value)
            metadata_group.create_dataset("material_name", data=material_name)
            metadata_group.create_dataset("distance_source_detector", data=distance_source_detector)
            metadata_group.create_dataset("detector_offset", data=detector_offset)
            metadata_group.create_dataset("bin_size", data=bin_size)

            # strain mapping dict
            strain_group = entry.create_group("strain mapping")
            for key in strain_mapping_dict.keys():
                key_group = strain_group.create_group(key)
                key_group.create_dataset("val", data=strain_mapping_dict[key]["val"])
                key_group.create_dataset("err", data=strain_mapping_dict[key]["err"])
                coordinate_group = key_group.create_group("bin coordinates")
                coordinate_group.create_dataset("x0", data=formatted_dict[key]["bin_coordinates"]["x0"])
                coordinate_group.create_dataset("y0", data=formatted_dict[key]["bin_coordinates"]["y0"])
                coordinate_group.create_dataset("x1", data=formatted_dict[key]["bin_coordinates"]["x1"])
                coordinate_group.create_dataset("y1", data=formatted_dict[key]["bin_coordinates"]["y1"])

            # fitting
            fitting_group = entry.create_group("fitting")

            # kropff
            kropff_group = fitting_group.create_group("kropff")

            for key in formatted_dict.keys():
                key_group = kropff_group.create_group(key)

                _item1 = formatted_dict[key]
                key_group.create_dataset("xaxis", data=_item1["xaxis"])
                key_group.create_dataset("yaxis", data=_item1["yaxis"])

                fitted_group = key_group.create_group("fitted")
                _item12 = _item1["fitted"]

                high_tof_group = fitted_group.create_group("high_tof")
                _item123 = _item12["high_tof"]
                if _item123["xaxis"]:
                    high_tof_group.create_dataset("xaxis", data=_item123["xaxis"])
                    high_tof_group.create_dataset("yaxis", data=_item123["yaxis"])
                else:
                    high_tof_group.create_dataset("xaxis", data="None")
                    high_tof_group.create_dataset("yaxis", data="None")

                low_tof_group = fitted_group.create_group("low_tof")
                _item123 = _item12["low_tof"]
                if _item123["xaxis"]:
                    low_tof_group.create_dataset("xaxis", data=_item123["xaxis"])
                    low_tof_group.create_dataset("yaxis", data=_item123["yaxis"])
                else:
                    low_tof_group.create_dataset("xaxis", data="None")
                    low_tof_group.create_dataset("yaxis", data="None")

                bragg_peak_group = fitted_group.create_group("bragg_peak")
                _item123 = _item12["bragg_peak"]
                if _item123["xaxis"]:
                    bragg_peak_group.create_dataset("xaxis", data=_item123["xaxis"])
                    bragg_peak_group.create_dataset("yaxis", data=_item123["yaxis"])
                else:
                    bragg_peak_group.create_dataset("xaxis", data="None")
                    bragg_peak_group.create_dataset("yaxis", data="None")

                for _item in [
                    "strain",
                    "d",
                    "a0",
                    "b0",
                    "ahkl",
                    "bhkl",
                    "tau",
                    "sigma",
                    "lambda_hkl",
                ]:
                    _group = fitted_group.create_group(_item)
                    _group.create_dataset("val", data=_item1[_item]["val"])
                    _group.create_dataset("err", data=_item1[_item]["err"])

                fitted_group.create_dataset("row_index", data=_item1[FittingKeys.row_index])
                fitted_group.create_dataset("column_index", data=_item1[FittingKeys.column_index])

                bragg_peak_threshold = fitted_group.create_group("bragg peak threshold")
                bragg_peak_threshold.create_dataset("left", data=_item1["bragg peak threshold"]["left"])
                bragg_peak_threshold.create_dataset("right", data=_item1["bragg peak threshold"]["right"])

            # integrated image
            integrated_image_group = entry.create_group("integrated normalized radiographs")
            integrated_image_group.create_dataset("2D array", data=integrated_image)

    def config_for_cli(self, output_folder: str = None):
        _current_time = get_current_time_in_special_file_name_format()

        output_folder = os.path.abspath(output_folder)
        output_file_name = create_full_export_file_name(
            os.path.join(output_folder, f"config_{_current_time}"), FileType.json
        )

        logging.info(f"Exporting configuration to be used by the command line version (CLI) -> {output_file_name}")

        # o_get = Get(parent=self.parent)
        # strain_mapping_dict = o_get.strain_mapping_dictionary()

        session_dict = self.grand_parent.session_dict

        raw_data_dir = session_dict[DataType.sample][SessionSubKeys.current_folder]
        if session_dict[DataType.sample][SessionSubKeys.list_files]:
            _, raw_data_extension = os.path.splitext(session_dict[DataType.sample][SessionSubKeys.list_files][0])
        else:
            raw_data_extension = None

        open_beam_data_dir = session_dict[DataType.ob][SessionSubKeys.current_folder]
        if session_dict[DataType.ob][SessionSubKeys.list_files]:
            _, open_beam_data_extension = os.path.splitext(session_dict[DataType.ob][SessionSubKeys.list_files][0])
        else:
            open_beam_data_extension = None

        # raw_data_dir = session_dict[DataType.normalized][SessionSubKeys.current_folder]
        # _, raw_data_extension = os.path.splitext(
        #     session_dict[DataType.normalized][SessionSubKeys.list_files][0]
        # )

        # open_beam_data_dir = session_dict[DataType.ob][SessionSubKeys.current_folder]
        # open_beam_data_extension = raw_data_extension

        list_normalization_roi = session_dict[DataType.normalization][SessionSubKeys.roi]
        normalization_sample_background = []
        for _roi in list_normalization_roi:
            _state, x0, y0, width, height, _type = _roi
            if (_type == "background") and _state:
                normalization_sample_background.append({"x0": x0, "y0": y0, "width": width, "height": height})

        moving_average_active = session_dict[SessionKeys.reduction][SessionSubKeys.activate]
        moving_average_dimension = session_dict[SessionKeys.reduction][SessionSubKeys.dimension]
        moving_average_size = {
            "y": session_dict[SessionKeys.reduction][SessionSubKeys.size]["y"],
            "x": session_dict[SessionKeys.reduction][SessionSubKeys.size]["x"],
        }
        if moving_average_dimension == "3D":
            moving_average_size["z"] = session_dict[SessionKeys.reduction][SessionSubKeys.size["z"]]

        moving_average_type = session_dict[SessionKeys.reduction][SessionSubKeys.type]

        if session_dict[SessionKeys.reduction][SessionSubKeys.process_order] == "option1":
            processing_order = "Moving average, Normalization"
        else:
            processing_order = "Normalization, Moving average"

        analysis_material_element = self.parent.ui.material_name.text()

        pixel_binning = {
            "x0": session_dict[SessionKeys.bin][SessionSubKeys.roi][1],
            "y0": session_dict[SessionKeys.bin][SessionSubKeys.roi][2],
            "width": session_dict[SessionKeys.bin][SessionSubKeys.roi][3],
            "height": session_dict[SessionKeys.bin][SessionSubKeys.roi][4],
            "bins_size": session_dict[SessionKeys.bin][SessionSubKeys.roi][5],
        }

        fitting_lambda_range = session_dict[SessionKeys.fitting][SessionSubKeys.lambda_range_index]
        x_axis = session_dict[SessionKeys.fitting][SessionSubKeys.xaxis]
        lambda_min = x_axis[fitting_lambda_range[0]] * 1e-10
        lambda_max = x_axis[fitting_lambda_range[1]] * 1e-10

        if self.parent.ui.d0_value.isChecked():
            strain_mapping_d0 = float(self.parent.ui.d0_value.text())
        else:
            strain_mapping_d0 = float(self.parent.ui.d0_user_value.text())

        strain_mapping_d0 = float(f"{strain_mapping_d0:04.4f}")

        quality_threshold = 0.8

        distance_source_detector_in_m = session_dict[SessionKeys.instrument][SessionSubKeys.distance_source_detector]
        detector_offset_in_us = session_dict[SessionKeys.instrument][SessionSubKeys.detector_value]

        normalized_data_dir = os.path.join(output_folder, f"normalized_{_current_time}")
        analysis_results_dir = os.path.join(output_folder, f"analysis_{_current_time}")
        strain_results_dir = os.path.join(output_folder, f"strain_{_current_time}")

        config = {
            "raw_data": {
                "raw_data_dir": raw_data_dir,
                "extension": raw_data_extension,
            },
            "open_beam": {
                "open_beam_data_dir": open_beam_data_dir,
                "extension": open_beam_data_extension,
            },
            "normalization": {
                "sample_background": normalization_sample_background,
                "moving_average": {
                    "active": moving_average_active,
                    "dimension": moving_average_dimension,
                    "size": moving_average_size,
                    "type": moving_average_type,
                },
                "processing_order": processing_order,
            },
            "analysis": {
                "material": {
                    "element": analysis_material_element,
                },
                "pixel_binning": pixel_binning,
                "fitting": {
                    "lambda_min": lambda_min,
                    "lambda_max": lambda_max,
                },
                "strain_mapping": {
                    "d0": strain_mapping_d0,
                    "quality_threshold": quality_threshold,
                    "visualization": {
                        "interpolation_method": "nearest",
                        "colormap": "viridis",
                        "alpha": 0.5,
                        "display_fit_quality": True,
                    },
                    "output_file_config": {
                        "strain_map_format": "png",
                        "fitting_grid_format": "pdf",
                        "figure_dpi": 300,
                        "csv_format": {
                            "delimiter": ",",
                            "include_metadata_header": True,
                            "metadata_comment_char": "#",
                            "na_rep": "null",
                        },
                    },
                    "save_intermediate_results": False,
                },
                "distance_source_detector_in_m": distance_source_detector_in_m,
                "detector_offset_in_us": detector_offset_in_us,
            },
            "output": {
                "normalized_data_dir": normalized_data_dir,
                "analysis_results_dir": analysis_results_dir,
                "strain_results_dir": strain_results_dir,
            },
        }

        save_json(json_file_name=output_file_name, json_dictionary=config)



from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from qtpy.QtWidgets import QSizePolicy


class MatplotlibView(FigureCanvas):
    def __init__(self, parent):
        self.fig = Figure()
        self.ax1 = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)



#!/usr/bin/env python
"""
KropffAutomaticThresholdAlgorithms class for handling the automatic threshold algorithms.
"""

import numpy as np
from changepy import pelt
from changepy.costs import normal_var
from qtpy import QtGui
from scipy.optimize import curve_fit
from scipy.special import erf

from ibeatles.fitting.kropff import KropffThresholdFinder


class MeanRangeCalculation(object):
    """
    Mean value of all the counts between left_pixel and right pixel
    """

    def __init__(self, data=None):
        self.data = data
        self.nbr_pixel = len(self.data)

    def calculate_left_right_mean(self, pixel=-1):
        _data = self.data
        _nbr_pixel = self.nbr_pixel

        self.left_mean = np.nanmean(_data[0 : pixel + 1])
        self.right_mean = np.nanmean(_data[pixel + 1 : _nbr_pixel])

    def calculate_delta_mean_square(self):
        self.delta_square = np.square(self.left_mean - self.right_mean)


class Algorithms:
    """This class calculates the front edge position of a set of profiles"""

    # dict_profiles = {}   # {'0': {'data': [], 'delta_time': 45455}, '1': {...} ...}

    # {'0': {'xaxis': [], 'yaxis': [], ... },
    #  '1': {'xaxis': [], 'yaxis': [], ... },
    #  '2': {'xaxis': [], 'yaxis': [], ... },
    #  }
    table_dictionary = None

    data_have_been_reversed_in_calculation = False

    peak_sliding_average_data = None

    # fitting by error function requires that the signal goes from max to min values
    is_data_from_max_to_min = True
    peak_error_function_data = None
    peak_error_function_data_error = None

    peak_change_point_data = None

    water_intake_peak_erf: list = []
    water_intake_deltatime: list = []

    dict_error_function_parameters: dict = {}

    progress_bar_ui = None  # progress bar ui

    def __init__(
        self,
        kropff_table_dictionary=None,
        algorithm_selected="sliding_average",
        progress_bar_ui=None,
    ):
        self.table_dictionary = kropff_table_dictionary
        self.progress_bar_ui = progress_bar_ui

        if algorithm_selected == KropffThresholdFinder.sliding_average:
            self.calculate_using_sliding_average()
        elif algorithm_selected == KropffThresholdFinder.error_function:
            self.calculate_using_erf()
        elif algorithm_selected == KropffThresholdFinder.change_point:
            self.calculate_change_point()
        else:
            raise ValueError("algorithm not implemented yet!")

    @staticmethod
    def fitting_function(x, c, w, m, n):
        return ((m - n) / 2.0) * erf((x - c) / w) + (m + n) / 2.0

    def are_data_from_max_to_min(self, ydata):
        nbr_points = len(ydata)
        nbr_point_for_investigation = int(nbr_points * 0.1)

        mean_first_part = np.mean(ydata[0:nbr_point_for_investigation])

        ydata = ydata.copy()
        ydata_reversed = ydata[::-1]
        mean_last_part = np.mean(ydata_reversed[0:nbr_point_for_investigation])

        if mean_first_part > mean_last_part:
            return True
        else:
            return False

    def calculate_change_point(self):
        table_dictionary = self.table_dictionary
        x_axis = table_dictionary["0"]["xaxis"]
        nbr_pixels = len(x_axis)
        nbr_files = len(table_dictionary.keys())

        _start_file = 0
        _end_file = nbr_files

        if self.progress_bar_ui:
            self.progress_bar_ui.setMaximum(nbr_files)
            self.progress_bar_ui.setVisible(True)
            QtGui.QGuiApplication.processEvents()

        water_intake_peaks = []
        for _index_file, _row in enumerate(table_dictionary.keys()):
            _profile_data = table_dictionary[_row]["yaxis"]
            var = np.mean(_profile_data)
            result = pelt(normal_var(_profile_data, var), nbr_pixels)
            if len(result) > 2:
                peak = int(np.mean(result[2:]))
            else:
                peak = int(np.mean(result[1:]))
            water_intake_peaks.append(peak)

            if self.progress_bar_ui:
                self.progress_bar_ui.setValue(_index_file + 1)
                QtGui.QGuiApplication.processEvents()

        self.peak_change_point_data = water_intake_peaks

        if self.progress_bar_ui:
            self.progress_bar_ui.setVisible(False)
            QtGui.QGuiApplication.processEvents()

    def calculate_using_erf(self):
        table_dictionary = self.table_dictionary

        nbr_files = len(table_dictionary.keys())

        _start_file = 0
        _end_file = nbr_files

        if self.progress_bar_ui:
            self.progress_bar_ui.setMaximum(nbr_files)
            self.progress_bar_ui.setVisible(True)
            QtGui.QGuiApplication.processEvents()

        # dict_error_function_parameters = dict()
        # water_intake_peaks_erf = []
        # water_intake_peaks_erf_error = []
        # delta_time = []

        peak_error_function_data = []
        peak_error_function_data_error = []
        dict_error_function_parameters = dict()
        for _index_file, _row in enumerate(table_dictionary.keys()):
            ydata = table_dictionary[_row]["yaxis"]
            is_data_from_max_to_min = self.are_data_from_max_to_min(ydata)
            self.is_data_from_max_to_min = is_data_from_max_to_min
            if not is_data_from_max_to_min:
                self.data_have_been_reversed_in_calculation = True
                ydata = ydata[::-1]

            (popt, pcov) = self.fitting_algorithm(ydata)

            _local_dict = {"c": popt[0], "w": popt[1], "m": popt[2], "n": popt[3]}

            error = np.sqrt(np.diag(pcov))
            _peak = int(popt[0] + (popt[1] / np.sqrt(2)))

            if not is_data_from_max_to_min:
                _peak = len(ydata) - _peak

            peak_error_function_data.append(_peak)

            for _i, _err in enumerate(error):
                if np.isnan(_err):
                    error[_i] = np.sqrt(popt[0])
                elif np.isinf(_err):
                    error[_i] = np.sqrt(popt[0])
                else:
                    error[_i] = _err

            _peak_error = int(error[0] + (error[1] / np.sqrt(2)))

            peak_error_function_data_error.append(_peak_error)
            dict_error_function_parameters[str(_index_file)] = _local_dict

            if self.progress_bar_ui:
                self.progress_bar_ui.setValue(_index_file + 1)
                QtGui.QGuiApplication.processEvents()

        self.peak_error_function_data = peak_error_function_data
        self.dict_error_function_parameters = dict_error_function_parameters
        self.peak_error_function_data_error = peak_error_function_data_error
        # self.water_intake_deltatime = delta_time

        if self.progress_bar_ui:
            self.progress_bar_ui.setVisible(False)
            QtGui.QGuiApplication.processEvents()

    def fitting_algorithm(self, ydata):
        fitting_xdata = np.arange(len(ydata))
        popt, pcov = curve_fit(self.fitting_function, fitting_xdata, ydata, maxfev=3000)
        return (popt, pcov)

    def calculate_using_sliding_average(self):
        table_dictionary = self.table_dictionary

        x_axis = table_dictionary["0"]["xaxis"]
        nbr_pixels = len(x_axis)
        nbr_files = len(table_dictionary.keys())

        _start_file = 0
        _end_file = nbr_files

        if self.progress_bar_ui:
            self.progress_bar_ui.setMaximum(nbr_files)
            self.progress_bar_ui.setVisible(True)
            QtGui.QGuiApplication.processEvents()

        peak_sliding_average_data = []
        for _index_file, _row in enumerate(table_dictionary.keys()):
            _profile_data = table_dictionary[_row]["yaxis"]
            delta_array = []
            _o_range = MeanRangeCalculation(data=_profile_data)
            for _pixel in np.arange(0, nbr_pixels):
                _o_range.calculate_left_right_mean(pixel=_pixel)
                _o_range.calculate_delta_mean_square()
                delta_array.append(_o_range.delta_square)

            peak_value = delta_array.index(max(delta_array[0:nbr_pixels]))
            peak_sliding_average_data.append(peak_value)

            if self.progress_bar_ui:
                self.progress_bar_ui.setValue(_index_file + 1)
                QtGui.QGuiApplication.processEvents()

        self.peak_sliding_average_data = peak_sliding_average_data
        if self.progress_bar_ui:
            self.progress_bar_ui.setVisible(False)
            QtGui.QGuiApplication.processEvents()

    def get_peak_value_array(self, algorithm_selected="sliding_average"):
        if algorithm_selected == KropffThresholdFinder.sliding_average:
            return self.peak_sliding_average_data
        elif algorithm_selected == KropffThresholdFinder.change_point:
            return self.peak_change_point_data
        elif algorithm_selected == KropffThresholdFinder.error_function:
            return self.peak_error_function_data
        else:
            raise ValueError("algorithm not implemented yet!")

    @staticmethod
    def bin_data(data=None, bin_size=1, bin_type="median"):
        numpy_data = np.array(data).flatten()
        if bin_size == 1:
            return numpy_data

        nbr_bin = int(len(numpy_data) / bin_size)
        data_to_rebinned = numpy_data[0 : nbr_bin * bin_size]
        binned_array_step1 = np.reshape(data_to_rebinned, [nbr_bin, bin_size])
        if bin_type == "mean":
            binned_array = np.mean(binned_array_step1, axis=1)
        elif bin_type == "median":
            binned_array = np.median(binned_array_step1, axis=1)
        else:
            raise NotImplementedError("bin data type not supported!")

        return binned_array



#!/usr/bin/env python
"""
Infos launcher
"""

import os

from qtpy.QtWidgets import QDialog

from ibeatles import DataType, load_ui
from ibeatles.utilities.gui_handler import GuiHandler


class InfosLauncher:
    def __init__(self, parent=None):
        self.parent = parent

        if self.parent.infos_id is None:
            infos_id = Infos(parent=self.parent)
            infos_id.show()
            self.parent.infos_id = infos_id
        else:
            self.parent.infos_id.activateWindow()
            self.parent.infos_id.setFocus()
            self.parent.infos_id.update()


class Infos(QDialog):
    def __init__(self, parent=None):
        self.parent = parent
        QDialog.__init__(self, parent=parent)
        ui_full_path = os.path.join(os.path.dirname(__file__), os.path.join("ui", "ui_infos.ui"))
        self.ui = load_ui(ui_full_path, baseinstance=self)
        self.setWindowTitle("Folders Infos")

        self.ui.sample_textEdit.setReadOnly(True)
        self.ui.ob_textEdit.setReadOnly(True)
        self.ui.normalized_textEdit.setReadOnly(True)
        self.update()

    def ok_clicked(self):
        self.close()

    def closeEvent(self, a0):
        self.parent.infos_id = None

    def update(self):
        # get current main tab activated (sample, ob or normalized)
        o_gui = GuiHandler(parent=self.parent)
        data_type = o_gui.get_active_tab()

        # jump to tab that corresponds to current main tab activated
        if data_type == DataType.sample:
            tab_index = 0
        elif data_type == DataType.ob:
            tab_index = 1
        else:
            tab_index = 2
        self.ui.toolBox.setCurrentIndex(tab_index)

        # disable or not tabs and add message if not
        infos_dict = self.parent.infos_dict
        if infos_dict[DataType.sample] is None:
            self.ui.toolBox.setItemEnabled(0, False)
        else:
            self.ui.toolBox.setItemEnabled(0, True)
            _entry = infos_dict[DataType.sample]
            text = ""
            for key in _entry:
                text += "<b>{}</b>: {}<br/>".format(_entry[key]["name"], _entry[key]["value"])

                self.ui.sample_textEdit.setHtml(text)

        if infos_dict[DataType.ob] is None:
            self.ui.toolBox.setItemEnabled(1, False)
        else:
            self.ui.toolBox.setItemEnabled(1, True)
            _entry = infos_dict[DataType.ob]
            text = ""
            for key in _entry:
                text += "<b>{}</b>: {}<br/>".format(_entry[key]["name"], _entry[key]["value"])

                self.ui.ob_textEdit.setHtml(text)

        if infos_dict[DataType.normalized] is None:
            self.ui.toolBox.setItemEnabled(2, False)
        else:
            self.ui.toolBox.setItemEnabled(2, True)
            _entry = infos_dict[DataType.normalized]
            text = ""
            for key in _entry:
                text += "<b>{}</b>: {}<br/>".format(_entry[key]["name"], _entry[key]["value"])

                self.ui.normalized_textEdit.setHtml(text)



class KropffTabSelected:
    settings = "initial settings"
    high_tof = "high_tof"
    low_tof = "low_tof"
    bragg_peak = "bragg_peak"
    summary = "summary"


class FittingTabSelected:
    march_dollase = "march_dollase"
    kropff = "kropff"


class FittingKeys:
    x_axis = "xaxis"
    y_axis = "yaxis"
    transparency = "transparency"
    image_view_state = "image view state"
    image_view_histogram = "image view histogram"
    ui_accessed = "ui accessed"
    ui = "ui"
    row_index = "row_index"
    column_index = "column_index"
    bin_coordinates = "bin_coordinates"
    active = "active"
    lock = "lock"


selected_color = {"pen": (0, 0, 0, 30), "brush": (0, 255, 0, 150)}

lock_color = {"pen": (0, 0, 0, 30), "brush": (255, 0, 0, 240)}



class TimeSpectraKeys:
    tof_array = "tof_array"
    lambda_array = "lambda_array"
    file_index_array = "file_index_array"
    file_name = "file_name"
    counts_array = "counts_array"


class CombineAlgorithm:
    """list of algorithm used to combine the folders"""

    mean = "mean"
    median = "median"



#!/usr/bin/env python
"""
Table Dictionary Handler
"""

import os

import numpy as np
import pandas as pd
from qtpy.QtWidgets import QFileDialog

from ibeatles.fitting import fitting_handler
from ibeatles.utilities.table_handler import TableHandler


class ColumnNameIndex:
    activate = 3
    lock = 2


class TableDictionaryHandler:
    selected_color = {"pen": (0, 0, 0, 30), "brush": (0, 255, 0, 150)}

    lock_color = {"pen": (0, 0, 0, 30), "brush": (255, 0, 0, 240)}

    # header = ['x0', 'y0', 'x1', 'y1', 'row_index', 'column_index', 'lock', 'active',
    #           'fitting_confidence', 'd_spacing_value', 'd_spacing_err', 'd_spacing_fixed',
    #           'sigma_value', 'sigma_err', '',
    #           'intensity_value', 'intensity_err', 'intensity_fixed',
    #           'alpha_value', 'alpha_err', 'alpha_fixed',
    #           'a1_value', 'a1_err', 'a1_fixed',
    #           'a2_value', 'a2_err', 'a2_fixed',
    #           'a5_value', 'a5_err', 'a5_fixed',
    #           'a6_value', 'a6_err', 'a6_fixed']

    def __init__(self, grand_parent=None, parent=None):
        self.grand_parent = grand_parent  # iBeatles main
        self.parent = parent  # fitting ui
        self.value_table_ui = self.parent.ui.value_table

    def fill_table_with_variable(self, variable_name="d_spacing", value=np.nan, list_keys=[], all_keys=False):
        table_dictionary = self.grand_parent.march_table_dictionary
        if all_keys:
            list_keys = table_dictionary.keys()

        for _key in list_keys:
            table_dictionary[_key][variable_name]["val"] = value

        self.grand_parent.march_table_dictionary = table_dictionary

    def populate_table_dictionary_entry(self, index=0, array=[]):
        table_dictionary = self.grand_parent.march_table_dictionary

        table_dictionary[str(index)] = {
            "bin_coordinates": {
                "x0": array[0],
                "x1": array[2],
                "y0": array[1],
                "y1": array[3],
            },
            "selected_item": None,
            "locked_item": None,
            "row_index": array[4],
            "column_index": array[5],
            "selected": False,
            "lock": array[6],
            "active": array[7],
            "rejected": False,  # use when bin is outside of the sample
            "fitting_confidence": array[8],
            "d_spacing": {"val": array[9], "err": array[10], "fixed": array[11]},
            "sigma": {"val": array[12], "err": array[13], "fixed": array[14]},
            "intensity": {"val": array[15], "err": array[16], "fixed": array[17]},
            "alpha": {"val": array[18], "err": array[19], "fixed": array[20]},
            "a1": {"val": array[21], "err": array[22], "fixed": array[23]},
            "a2": {"val": array[24], "err": array[25], "fixed": array[26]},
            "a5": {"val": array[27], "err": array[28], "fixed": array[29]},
            "a6": {"val": array[30], "err": array[31], "fixed": array[32]},
        }

        self.grand_parent.march_table_dictionary = table_dictionary

    # def initialize_parameters_from_session(self):
    #     session_table_dictionary = self.grand_parent.session_dict["fitting"]['march dollase']["table dictionary"]
    #     table_dictionary = self.grand_parent.march_table_dictionary
    #
    #     for _row in session_table_dictionary.keys():
    #         _entry = session_table_dictionary[_row]
    #         table_dictionary[_row]['lock'] = _entry['lock']
    #         table_dictionary[_row]['active'] = _entry['active']
    #         table_dictionary[_row]['fitting_confidence'] = _entry['fitting_confidence']
    #         table_dictionary[_row]['d_spacing'] = _entry['d_spacing']
    #         table_dictionary[_row]['sigma'] = _entry['sigma']
    #         table_dictionary[_row]['alpha'] = _entry['alpha']
    #         table_dictionary[_row]['a1'] = _entry['a1']
    #         table_dictionary[_row]['a2'] = _entry['a2']
    #         table_dictionary[_row]['a5'] = _entry['a5']
    #         table_dictionary[_row]['a6'] = _entry['a6']
    #
    #     lambda_range = self.grand_parent.session_dict["fitting"]["lambda range index"]
    #     if lambda_range:
    #         [lambda_min_index, lambda_max_index] = self.grand_parent.session_dict["fitting"]["lambda range index"]
    #         x_axis = self.grand_parent.session_dict["fitting"]["x_axis"]
    #
    #         lambda_min = x_axis[lambda_min_index]
    #         lambda_max = x_axis[lambda_max_index]
    #
    #         self.parent.ui.lambda_min_lineEdit.setText("{:4.2f}".format(lambda_min))
    #         self.parent.ui.lambda_max_lineEdit.setText("{:4.2f}".format(lambda_max))
    #         self.grand_parent.fitting_bragg_edge_linear_selection = [lambda_min_index, lambda_max_index]
    #
    #     transparency = self.grand_parent.session_dict['fitting']['transparency']
    #     self.parent.ui.slider.setValue(transparency)
    #
    #     self.grand_parent.display_active_row_flag = \
    #         self.grand_parent.session_dict['fitting']['march dollase']['plot active row flag']
    #     self.parent.ui.active_bins_button.setChecked(self.grand_parent.display_active_row_flag)
    #     self.parent.ui.locked_bins_button.setChecked(not self.grand_parent.display_active_row_flag)
    #
    #     self.grand_parent.march_table_dictionary = table_dictionary
    #     self.grand_parent.table_loaded_from_session = None

    def clear_y_axis_and_x_axis_from_kropff_table_dictionary(self):
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        for _row in kropff_table_dictionary.keys():
            kropff_table_dictionary[_row]["yaxis"] = None
            kropff_table_dictionary[_row]["xaxis"] = None

    # def create_table_dictionary(self):
    #     '''
    #     this will define the corner position and index of each cell
    #     '''
    #     # if len(np.array(self.grand_parent.data_metadata['normalized']['data_live_selection'])) == 0:
    #     #     return
    #     #
    #     # if not self.grand_parent.march_table_dictionary == {}:
    #     #     return
    #
    #     bin_size = self.grand_parent.binning_roi[-1]
    #     pos = self.grand_parent.binning_line_view['pos']
    #
    #     # calculate outside real edges of bins
    #     min_max_xy = get_min_max_xy(pos)
    #
    #     from_x = min_max_xy['x']['min']
    #     to_x = min_max_xy['x']['max']
    #
    #     from_y = min_max_xy['y']['min']
    #     to_y = min_max_xy['y']['max']
    #
    #     march_table_dictionary = {}
    #     kropff_table_dictionary = {}
    #     _index = 0
    #     _index_col = 0
    #     for _x in np.arange(from_x, to_x, bin_size):
    #         _index_row = 0
    #         for _y in np.arange(from_y, to_y, bin_size):
    #             _str_index = str(_index)
    #
    #             kropff_table_dictionary[_str_index] = {'bin_coordinates': {'x0': _x,
    #                                                                        'x1': _x + bin_size,
    #                                                                        'y0': _y,
    #                                                                        'y1': _y + bin_size},
    #                                                    'yaxis': None,
    #                                                    'xaxis': None,
    #                                                    'selected_item': None,
    #                                                    'locked_item': None,
    #                                                    'row_index': _index_row,
    #                                                    'column_index': _index_col,
    #                                                    'selected': False,
    #                                                    'lock': False,
    #                                                    'active': False,
    #                                                    'a0': {'val': np.nan,
    #                                                           'err': np.nan},
    #                                                    'b0': {'val': np.nan,
    #                                                           'err': np.nan},
    #                                                    'ahkl': {'val': np.nan,
    #                                                             'err': np.nan},
    #                                                    'bhkl': {'val': np.nan,
    #                                                             'err': np.nan},
    #                                                    'lambda_hkl': {'val': np.nan,
    #                                                                   'err': np.nan},
    #                                                    'tau': {'val': np.nan,
    #                                                            'err': np.nan},
    #                                                    'sigma': {'val': np.nan,
    #                                                              'err': np.nan},
    #                                                    'bragg peak threshold': {'left': np.nan,
    #                                                                             'right': np.nan},
    #                                                    }
    #
    #             # create the box to show when bin is selected
    #             selection_box = pg.QtGui.QGraphicsRectItem(_x, _y,
    #                                                        bin_size,
    #                                                        bin_size)
    #             selection_box.setPen(pg.mkPen(self.selected_color['pen']))
    #             selection_box.setBrush(pg.mkBrush(self.selected_color['brush']))
    #             kropff_table_dictionary[_str_index]['selected_item'] = selection_box
    #
    #             march_table_dictionary[_str_index] = {'bin_coordinates': {'x0': _x,
    #                                                                       'x1': _x + bin_size,
    #                                                                       'y0': _y,
    #                                                                       'y1': _y + bin_size},
    #                                                    'selected_item': None,
    #                                                    'locked_item': None,
    #                                                    'row_index': _index_row,
    #                                                    'column_index': _index_col,
    #                                                    'selected': False,
    #                                                    'lock': False,
    #                                                    'active': False,
    #                                                    'fitting_confidence': np.nan,
    #                                                    'd_spacing': {'val': np.nan,
    #                                                                  'err': np.nan,
    #                                                                  'fixed': False},
    #                                                    'sigma': {'val': np.nan,
    #                                                              'err': np.nan,
    #                                                              'fixed': False},
    #                                                    'intensity': {'val': np.nan,
    #                                                                  'err': np.nan,
    #                                                                  'fixed': False},
    #                                                    'alpha': {'val': np.nan,
    #                                                              'err': np.nan,
    #                                                              'fixed': False},
    #                                                    'a1': {'val': np.nan,
    #                                                           'err': np.nan,
    #                                                           'fixed': False},
    #                                                    'a2': {'val': np.nan,
    #                                                           'err': np.nan,
    #                                                           'fixed': False},
    #                                                    'a5': {'val': np.nan,
    #                                                           'err': np.nan,
    #                                                           'fixed': False},
    #                                                    'a6': {'val': np.nan,
    #                                                           'err': np.nan,
    #                                                           'fixed': False},
    #                                                   }
    #
    #             # march_table_dictionary[_str_index]['bin_coordinates']['x0'] = _x
    #             # march_table_dictionary[_str_index]['bin_coordinates']['x1'] = _x + bin_size
    #             # march_table_dictionary[_str_index]['bin_coordinates']['y0'] = _y
    #             # march_table_dictionary[_str_index]['bin_coordinates']['y1'] = _y + bin_size
    #
    #             # create the box to show when bin is selected
    #             selection_box = pg.QtGui.QGraphicsRectItem(_x, _y,
    #                                                        bin_size,
    #                                                        bin_size)
    #             selection_box.setPen(pg.mkPen(self.selected_color['pen']))
    #             selection_box.setBrush(pg.mkBrush(self.selected_color['brush']))
    #             march_table_dictionary[_str_index]['selected_item'] = selection_box
    #
    #             # create the box to show when bin is locked
    #             lock_box = pg.QtGui.QGraphicsRectItem(_x, _y,
    #                                                   bin_size,
    #                                                   bin_size)
    #             lock_box.setPen(pg.mkPen(self.lock_color['pen']))
    #             lock_box.setBrush(pg.mkBrush(self.lock_color['brush']))
    #             march_table_dictionary[_str_index]['locked_item'] = lock_box
    #
    #             _index += 1
    #             _index_row += 1
    #
    #         _index_col += 1
    #
    #     self.grand_parent.march_table_dictionary = march_table_dictionary
    #     self.grand_parent.kropff_table_dictionary = kropff_table_dictionary
    #
    #     self.grand_parent.fitting_selection['nbr_row'] = _index_row
    #     self.grand_parent.fitting_selection['nbr_column'] = _index_col

    def full_table_selection_tool(self, status=True):
        o_table = TableHandler(table_ui=self.value_table_ui)
        o_table.select_everything(status)

    def unselect_full_table(self):
        self.full_table_selection_tool(status=False)

    def select_full_table(self):
        self.full_table_selection_tool(status=True)

    # def get_average_parameters_activated(self):
    #     table_dictionary = self.grand_parent.march_table_dictionary
    #
    #     d_spacing = []
    #     alpha = []
    #     sigma = []
    #     a1 = []
    #     a2 = []
    #     a5 = []
    #     a6 = []
    #
    #     for _index in table_dictionary.keys():
    #         _entry = table_dictionary[_index]
    #
    #         if _entry['active']:
    #             _d_spacing = _entry['d_spacing']['val']
    #             _alpha = _entry['alpha']['val']
    #             _sigma = _entry['sigma']['val']
    #             _a1 = _entry['a1']['val']
    #             _a2 = _entry['a2']['val']
    #             _a5 = _entry['a5']['val']
    #             _a6 = _entry['a6']['val']
    #
    #             d_spacing.append(_d_spacing)
    #             alpha.append(_alpha)
    #             sigma.append(_sigma)
    #             a1.append(_a1)
    #             a2.append(_a2)
    #             a5.append(_a5)
    #             a6.append(_a6)
    #
    #     mean_d_spacing = self.get_mean_value(d_spacing)
    #     mean_alpha = self.get_mean_value(alpha)
    #     mean_sigma = self.get_mean_value(sigma)
    #     mean_a1 = self.get_mean_value(a1)
    #     mean_a2 = self.get_mean_value(a2)
    #     mean_a5 = self.get_mean_value(a5)
    #     mean_a6 = self.get_mean_value(a6)
    #
    #     return {'d_spacing': mean_d_spacing,
    #             'alpha': mean_alpha,
    #             'sigma': mean_sigma,
    #             'a1': mean_a1,
    #             'a2': mean_a2,
    #             'a5': mean_a5,
    #             'a6': mean_a6}

    # def get_mean_value(self, array=[]):
    #     if array == []:
    #         return np.nan
    #     else:
    #         return np.mean(array)

    def import_table(self):
        default_file_name = str(self.grand_parent.ui.normalized_folder.text()) + "_fitting_table.csv"
        table_file = str(
            QFileDialog.getOpenFileName(
                self.grand_parent,
                "Define Location and File Name Where to Export the Table!",
                os.path.join(self.grand_parent.normalized_folder, default_file_name),
            )
        )

        if table_file:
            pandas_data_frame = pd.read_csv(table_file)
            o_table = TableDictionaryHandler(grand_parent=self.grand_parent)

            numpy_table = pandas_data_frame.values
            # loop over each row in the pandas data frame
            for _index, _row_values in enumerate(numpy_table):
                o_table.populate_table_dictionary_entry(index=_index, array=_row_values)

            o_fitting = fitting_handler.FittingHandler(grand_parent=self.grand_parent)
            o_fitting.fill_table()

    def is_at_least_one_row_activated(self):
        nbr_row = self.value_table_ui.rowCount()
        for _row in np.arange(nbr_row):
            if self.is_this_row_checked(row=_row, column=ColumnNameIndex.activate):
                return True
        return False

    def is_this_row_checked(self, row=0, column=2):
        widget = self.parent.ui.value_table.cellWidget(row, column)
        return widget.isChecked()



from enum import Enum

import numpy as np
from scipy.constants import c, electron_volt, h, m_n

""" Small library of units and conversions for use in SAMMY fitting. """


class TimeUnitOptions(str, Enum):
    s = "s"
    ms = "ms"
    us = "us"
    ns = "ns"
    ps = "ps"


class EnergyUnitOptions(str, Enum):
    meV = "meV"
    eV = "eV"
    keV = "keV"
    MeV = "MeV"
    GeV = "GeV"
    J = "J"


class CrossSectionUnitOptions(str, Enum):
    barn = "barn"
    millibarn = "millibarn"
    microbarn = "microbarn"
    cm2 = "cm^2"


class DistanceUnitOptions(str, Enum):
    cm = "cm"
    nm = "nm"
    pm = "pm"
    m = "m"
    angstrom = "angstrom"


def convert_time_units(from_unit, to_unit):
    """Convert time from one unit to another unit
    based on TimeUnitOptions options

    Args:
        from_unit (TimeUnitOptions): Unit to convert from.
        to_unit (TimeUnitOptions): Unit to convert to.

    Returns:
        float: Time in the new unit.
    """

    # Conversion factors
    conversion_factors = {
        TimeUnitOptions.s: 1,
        TimeUnitOptions.ms: 1e-3,
        TimeUnitOptions.us: 1e-6,
        TimeUnitOptions.ns: 1e-9,
        TimeUnitOptions.ps: 1e-12,
    }

    return conversion_factors[from_unit] / conversion_factors[to_unit]


def convert_distance_units(from_unit, to_unit):
    """Convert distance from one unit to another unit
    based on DistanceUnitOptions options

    Args:
        from_unit (DistanceUnitOptions): Unit to convert from.
        to_unit (DistanceUnitOptions): Unit to convert to.

    Returns:
        float: distance in the new unit.
    """

    # Conversion factors
    conversion_factors = {
        DistanceUnitOptions.nm: 1e-9,
        DistanceUnitOptions.cm: 1e-2,
        DistanceUnitOptions.pm: 1e-12,
        DistanceUnitOptions.m: 1,
        DistanceUnitOptions.angstrom: 1e-10,
    }

    return conversion_factors[from_unit] / conversion_factors[to_unit]


def convert_to_energy(from_unit, to_unit):
    """Convert energy from one unit to another unit
    based on EnergyUnitOptions options

    Args:
        from_unit (EnergyUnitOptions): Unit to convert from.
        to_unit (EnergyUnitOptions): Unit to convert to.

    Returns:
        float: Energy in the new unit.
    """

    # Conversion factors
    conversion_factors = {
        EnergyUnitOptions.meV: 1e-3,
        EnergyUnitOptions.eV: 1,
        EnergyUnitOptions.keV: 1e3,
        EnergyUnitOptions.MeV: 1e6,
        EnergyUnitOptions.GeV: 1e9,
        EnergyUnitOptions.J: 1 / electron_volt,
    }

    return conversion_factors[from_unit] / conversion_factors[to_unit]


def convert_to_cross_section(from_unit, to_unit):
    """Convert cross section from one unit to another unit
    based on CrossSectionUnitOptions options

    Args:
        from_unit (CrossSectionUnitOptions): Unit to convert from.
        to_unit (CrossSectionUnitOptions): Unit to convert to.

    Returns:
        float: Cross section in the new unit.
    """

    # Conversion factors
    conversion_factors = {
        CrossSectionUnitOptions.barn: 1,
        CrossSectionUnitOptions.millibarn: 1e-3,
        CrossSectionUnitOptions.microbarn: 1e-6,
        CrossSectionUnitOptions.cm2: 1e-24,
    }

    return conversion_factors[from_unit] / conversion_factors[to_unit]


def convert_from_wavelength_to_energy_ev(wavelength, unit_from=DistanceUnitOptions.angstrom):
    """Convert wavelength to energy based on the given units.

    Args:
        wavelength (float): Wavelength value.
        unit_from (DistanceUnitOptions): Unit of the input wavelength.

    Returns:
        float: Energy in eV.
    """
    wavelength_m = wavelength * convert_distance_units(unit_from, DistanceUnitOptions.m)
    energy = h * c / wavelength_m
    energy = energy * convert_to_energy(EnergyUnitOptions.J, EnergyUnitOptions.eV)
    return energy


def convert_array_from_time_to_lambda(
    time_array: np.ndarray,
    time_unit: TimeUnitOptions,
    distance_source_detector: float,
    distance_source_detector_unit: DistanceUnitOptions,
    lambda_unit: DistanceUnitOptions,
    detector_offset_unit: DistanceUnitOptions,
    detector_offset: float = 0.0,
) -> np.ndarray:
    """Convert an array of time values to wavelength values.

    Args:
        time_array (np.ndarray): Array of time values.
        time_unit (TimeUnitOptions): Unit of the input time.
        distance_source_detector (float): Distance from the source to the detector.
        distance_source_detector_unit (DistanceUnitOptions): Unit of the distance.
        detector_offset (float): Offset of the detector.
        detector_offset_unit (DistanceUnitOptions): Unit of the offset.
        lambda_unit (DistanceUnitOptions): Unit of the output wavelength.

    This is using the formula: lambda_m = h/(m_n * distance_source_detector_m) * (time_array_s + detector_offset_s)

    Returns:
        np.ndarray: Array of wavelength values.
    """
    time_array_s = time_array * convert_time_units(time_unit, TimeUnitOptions.s)
    detector_offset_s = detector_offset * convert_time_units(detector_offset_unit, TimeUnitOptions.s)
    distance_source_detector_m = distance_source_detector * convert_distance_units(
        distance_source_detector_unit, DistanceUnitOptions.m
    )

    h_over_mn = h / m_n
    lambda_m = h_over_mn * (time_array_s + detector_offset_s) / distance_source_detector_m

    lambda_converted = lambda_m * convert_distance_units(DistanceUnitOptions.m, lambda_unit)

    return lambda_converted


def convert_array_from_time_to_energy(
    time_array: np.ndarray,
    time_unit: TimeUnitOptions,
    distance_source_detector: float,
    distance_source_detector_unit: DistanceUnitOptions,
    detector_offset: float,
    detector_offset_unit: DistanceUnitOptions,
    energy_unit: EnergyUnitOptions,
) -> np.ndarray:
    """Convert an array of time values to energy values.

    Args:
        time_array (np.ndarray): Array of time values.
        time_unit (TimeUnitOptions): Unit of the input time.
        distance_source_detector (float): Distance from the source to the detector.
        distance_source_detector_unit (DistanceUnitOptions): Unit of the distance.
        detector_offset (float): Offset of the detector.
        detector_offset_unit (DistanceUnitOptions): Unit of the offset.
        energy_unit (EnergyUnitOptions): Unit of the output energy.

    This is using the formula: E_ev = 1/2 m_n v^2
    with t_tof = L/v (L distance source to detector in m, v velocity of the neutron in m/s)
    so E_ev = 1/2 m_n (L/t_tof)^2

    Returns:
        np.ndarray: Array of energy values.
    """

    time_units_factor = convert_time_units(time_unit, TimeUnitOptions.s)
    time_array_s = time_array * time_units_factor

    detector_units_factor = convert_time_units(detector_offset_unit, TimeUnitOptions.s)
    detector_offset = detector_units_factor * detector_offset

    distance_source_detector_factor = convert_distance_units(distance_source_detector_unit, DistanceUnitOptions.m)
    distance_source_detector_m = distance_source_detector * distance_source_detector_factor

    # Calculate the energy in eV using the formula E_ev = 1/2 m_n (L/t_tof)^2 / electron_volt

    full_time_array_s = time_array_s + detector_offset
    energy_array_ev = 0.5 * m_n * (distance_source_detector_m / full_time_array_s) ** 2 / electron_volt

    energy_array_factor = convert_to_energy(EnergyUnitOptions.eV, energy_unit)
    energy_array = energy_array_ev * energy_array_factor
    energy_array = np.array(energy_array)

    return energy_array
